# Generated by Django 2.1.7 on 2019-03-09 20:43

from decimal import Decimal

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0004_auto_20190309_1004"),
        ("services", "0002_auto_20190304_2247"),
        ("invoices", "0005_auto_20190307_0925"),
    ]

    operations = [
        migrations.CreateModel(
            name="Cost",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "cost",
                    models.DecimalField(
                        decimal_places=2,
                        default=None,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="cost",
                    ),
                ),
                (
                    "third_party_costs",
                    models.DecimalField(
                        blank=True,
                        decimal_places=2,
                        default=None,
                        help_text="Total incl. tax for third-party services.",
                        max_digits=10,
                        null=True,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="third party costs",
                    ),
                ),
            ],
            options={
                "verbose_name": "cost",
                "verbose_name_plural": "costs",
                "ordering": ["pk"],
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Effort",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "billing_per_hour",
                    models.DecimalField(
                        decimal_places=2,
                        default=None,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="billing per hour",
                    ),
                ),
                (
                    "hours",
                    models.DecimalField(
                        decimal_places=1,
                        max_digits=4,
                        validators=[
                            django.core.validators.MinValueValidator(Decimal("0.1"))
                        ],
                        verbose_name="hours",
                    ),
                ),
            ],
            options={
                "verbose_name": "effort",
                "verbose_name_plural": "efforts",
                "ordering": ["pk"],
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Service",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
                (
                    "effort_hours",
                    models.DecimalField(
                        decimal_places=1,
                        default=0,
                        max_digits=4,
                        validators=[
                            django.core.validators.MinValueValidator(Decimal("0.1"))
                        ],
                        verbose_name="effort hours",
                    ),
                ),
                (
                    "cost",
                    models.DecimalField(
                        decimal_places=2,
                        default=0,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="cost",
                    ),
                ),
                (
                    "invoice",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="services",
                        to="invoices.Invoice",
                        verbose_name="invoice",
                    ),
                ),
                (
                    "project_service",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="invoice_services",
                        to="projects.Service",
                        verbose_name="project service",
                    ),
                ),
            ],
            options={
                "verbose_name": "service",
                "verbose_name_plural": "services",
                "ordering": ["position", "created_at"],
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="effort",
            name="service",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="efforts",
                to="invoices.Service",
                verbose_name="service",
            ),
        ),
        migrations.AddField(
            model_name="effort",
            name="service_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="services.ServiceType",
                verbose_name="service type",
            ),
        ),
        migrations.AddField(
            model_name="cost",
            name="service",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="costs",
                to="invoices.Service",
                verbose_name="service",
            ),
        ),
    ]



from collections import defaultdict

from django.db import models
from django.utils.text import capfirst
from django.utils.translation import gettext, gettext_lazy as _


class Circle(models.Model):
    name = models.CharField(_("name"), max_length=100)

    class Meta:
        ordering = ["name"]
        verbose_name = _("circle")
        verbose_name_plural = _("circles")

    def __str__(self):
        return self.name


class RoleQuerySet(models.QuerySet):
    def choices(self):
        circles = defaultdict(list)
        for role in self.filter(is_removed=False).select_related("circle"):
            circles[role.circle].append(role)
        return [("", "----------")] + [
            (circle.name, [(role.id, str(role)) for role in roles])
            for circle, roles in sorted(circles.items(), key=lambda row: row[0].name)
        ]


class Role(models.Model):
    PAID_WORK = "paid-work"
    KNOWLEDGE_TRANSFER = "knowledge-transfer"
    SOCIAL_CARE = "social-care"
    OUTREACH = "outreach"
    OTHER = "other"

    CATEGORIES = [
        (
            PAID_WORK,
            _("Paid work"),
            "",
        ),
        (
            KNOWLEDGE_TRANSFER,
            _("Knowledge transfer"),
            _("debriefings, reading, exchanges"),
        ),
        (
            SOCIAL_CARE,
            _("Social care"),
            _("personal development, people care, care for the work environment"),
        ),
        (
            OUTREACH,
            _("Outreach"),
            _("corporate communication, acquisition"),
        ),
        (
            OTHER,
            _("Other internal matters"),
            _("administration, bookkeeping, coordination, etc."),
        ),
    ]

    CATEGORY_CHOICES = [row[:2] for row in CATEGORIES]
    CATEGORY_DESCRIPTION = {row[0]: row[2] for row in CATEGORIES}

    circle = models.ForeignKey(
        Circle, on_delete=models.CASCADE, related_name="roles", verbose_name=_("circle")
    )
    name = models.CharField(_("name"), max_length=100)
    for_circle = models.BooleanField(_("for the circle"), default=False)
    is_removed = models.BooleanField(_("is removed"), default=False)
    work_category = models.CharField(
        _("category"),
        max_length=20,
        choices=CATEGORY_CHOICES,
        blank=True,
    )

    objects = RoleQuerySet.as_manager()

    class Meta:
        ordering = ["is_removed", "-for_circle", "name"]
        verbose_name = _("role")
        verbose_name_plural = _("roles")

    def __str__(self):
        prefix = gettext("(removed)") + " " if self.is_removed else ""
        if self.for_circle:
            return "%s%s [%s]" % (
                prefix,
                capfirst(gettext("for the circle")),
                self.circle,
            )
        return "%s%s [%s]" % (prefix, self.name, self.circle)

    @property
    def pretty_name(self):
        return capfirst(gettext("for the circle")) if self.for_circle else self.name

    def get_work_category_description(self):
        return self.CATEGORY_DESCRIPTION.get(self.work_category, self.work_category)



# Generated by Django 2.1.7 on 2019-03-10 10:34

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0006_auto_20190309_2143"),
        ("logbook", "0003_auto_20190304_2247"),
    ]

    operations = [
        migrations.RemoveField(model_name="loggedcost", name="invoice"),
        migrations.RemoveField(model_name="loggedhours", name="invoice"),
        migrations.AddField(
            model_name="loggedcost",
            name="invoice_service",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="invoices.Service",
                verbose_name="invoice service",
            ),
        ),
        migrations.AddField(
            model_name="loggedhours",
            name="invoice_service",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="invoices.Service",
                verbose_name="invoice service",
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-15 10:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("services", "0003_auto_20190312_1535")]

    operations = [
        migrations.AlterField(
            model_name="servicetype",
            name="position",
            field=models.IntegerField(default=0, verbose_name="position"),
        )
    ]



# Generated by Django 3.0.4 on 2020-03-28 16:51

import datetime

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("logbook", "0016_auto_20191210_1537"),
    ]

    operations = [
        migrations.CreateModel(
            name="Break",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "day",
                    models.DateField(default=datetime.date.today, verbose_name="day"),
                ),
                ("starts_at", models.TimeField(verbose_name="starts at")),
                ("ends_at", models.TimeField(verbose_name="ends at")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="breaks",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "break",
                "verbose_name_plural": "breaks",
                "ordering": ["-day", "-starts_at"],
            },
        ),
        migrations.RunSQL("SELECT audit_audit_table('logbook_break')"),
    ]



import datetime as dt
from collections import defaultdict

from django.db.models import Sum
from django.utils import timezone

from workbench.invoices.models import Invoice
from workbench.logbook.models import LoggedCost, LoggedHours
from workbench.offers.models import Offer
from workbench.tools.formats import Z1, Z2


def project_budget_statistics(projects, *, cutoff_date=None):
    cutoff_date = cutoff_date or dt.date.today()
    cutoff_dttm = timezone.make_aware(dt.datetime.combine(cutoff_date, dt.time.max))

    costs = (
        LoggedCost.objects.filter(
            service__project__in=projects, rendered_on__lte=cutoff_dttm
        )
        .order_by()
        .values("service__project")
    )

    cost_per_project = {
        row["service__project"]: row["cost__sum"] for row in costs.annotate(Sum("cost"))
    }
    third_party_costs_per_project = {
        row["service__project"]: row["third_party_costs__sum"]
        for row in costs.filter(
            third_party_costs__isnull=False, invoice_service__isnull=True
        ).annotate(Sum("third_party_costs"))
    }

    hours = (
        LoggedHours.objects.filter(
            service__project__in=projects, rendered_on__lte=cutoff_dttm
        )
        .order_by()
        .values("service__project", "service__effort_rate")
        .annotate(Sum("hours"))
    )
    effort_cost_per_project = defaultdict(lambda: Z2)
    effort_hours_with_rate_undefined_per_project = defaultdict(lambda: Z1)
    hours_per_project = defaultdict(lambda: Z1)

    for row in hours:
        if row["service__effort_rate"] is None:
            effort_hours_with_rate_undefined_per_project[
                row["service__project"]
            ] += row["hours__sum"]
        else:
            effort_cost_per_project[row["service__project"]] += (
                row["service__effort_rate"] * row["hours__sum"]
            )
        hours_per_project[row["service__project"]] += row["hours__sum"]

    not_archived_hours = {
        row["service__project"]: row["hours__sum"]
        for row in hours.filter(archived_at__isnull=True)
        .values("service__project")
        .annotate(Sum("hours"))
    }

    offered_per_project = {
        row["project"]: row["total_excl_tax__sum"]
        for row in Offer.objects.accepted()
        .filter(project__in=projects)
        .order_by()
        .values("project")
        .annotate(Sum("total_excl_tax"))
    }
    invoiced_per_project = {
        row["project"]: row["total_excl_tax__sum"]
        for row in Invoice.objects.invoiced()
        .filter(project__in=projects, invoiced_on__lte=cutoff_date)
        .order_by()
        .values("project")
        .annotate(Sum("total_excl_tax"))
    }

    statistics = [
        {
            "project": project,
            "logbook": cost_per_project.get(project.id, Z2)
            + effort_cost_per_project[project.id],
            "cost": cost_per_project.get(project.id, Z2),
            "effort_cost": effort_cost_per_project[project.id],
            "effort_hours_with_rate_undefined": effort_hours_with_rate_undefined_per_project[  # noqa
                project.id
            ],
            "third_party_costs": third_party_costs_per_project.get(project.id, Z2),
            "offered": offered_per_project.get(project.id, Z2),
            "invoiced": invoiced_per_project.get(project.id, Z2),
            "hours": hours_per_project[project.id],
            "not_archived": not_archived_hours.get(project.id, Z1),
            "delta": cost_per_project.get(project.id, Z2)
            + effort_cost_per_project[project.id]
            - invoiced_per_project.get(project.id, Z2),
        }
        for project in projects
    ]
    overall = {
        key: sum(s[key] for s in statistics)
        for key in [
            "logbook",
            "cost",
            "effort_cost",
            "effort_hours_with_rate_undefined",
            "third_party_costs",
            "offered",
            "invoiced",
            "hours",
            "not_archived",
        ]
    }
    overall["delta_positive"] = sum(s["delta"] for s in statistics if s["delta"] > 0)
    overall["delta_negative"] = sum(s["delta"] for s in statistics if s["delta"] < 0)

    return {
        "statistics": sorted(statistics, key=lambda s: s["delta"], reverse=True),
        "overall": overall,
    }



import datetime as dt
from itertools import islice

from django import forms
from django.db.models import Q, Sum
from django.utils.dateparse import parse_date
from django.utils.html import format_html, format_html_join
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import Team, User
from workbench.invoices.utils import recurring
from workbench.planning.models import PlannedWork, PlanningRequest
from workbench.projects.models import Project
from workbench.tools.formats import Z1, local_date_format
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea, add_prefix
from workbench.tools.validation import monday


class PlanningRequestSearchForm(Form):
    project = forms.ModelChoiceField(
        queryset=Project.objects.all(),
        required=False,
        widget=Autocomplete(model=Project),
        label="",
    )
    created_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["created_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        if data.get("project"):
            queryset = queryset.filter(project=data.get("project"))
        queryset = self.apply_owned_by(queryset, attribute="created_by")
        return queryset.select_related(
            "created_by",
            "project__owned_by",
            "project__customer",
            "project__contact__organization",
        )


@add_prefix("modal")
class PlanningRequestForm(ModelForm):
    user_fields = default_to_current_user = ("created_by",)

    class Meta:
        model = PlanningRequest
        fields = (
            "offer",
            "title",
            "description",
            "is_provisional",
            "requested_hours",
            "earliest_start_on",
            "completion_requested_on",
            "receivers",
        )
        widgets = {
            "description": Textarea,
            "receivers": forms.CheckboxSelectMultiple,
        }

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        if self.project:  # Creating a new object
            kwargs.setdefault("initial", {}).update({"title": self.project.title})
        else:
            self.project = kwargs["instance"].project

        initial = kwargs.setdefault("initial", {})
        request = kwargs["request"]

        if offer_id := request.GET.get("plan_offer"):
            try:
                offer = self.project.offers.get(pk=offer_id)
            except (self.project.offers.model.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "title": offer.title,
                        "description": offer.description,
                        "requested_hours": offer.services.aggregate(
                            h=Sum("service_hours")
                        )["h"]
                        or Z1,
                    }
                )

        if service_id := request.GET.get("service"):
            try:
                service = self.project.services.get(pk=service_id)
            except (self.project.services.model.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "title": f"{self.project.title}: {service.title}",
                        "description": service.description,
                        "requested_hours": service.service_hours,
                    }
                )

        super().__init__(*args, **kwargs)
        self.instance.project = self.project

        self.fields[
            "offer"
        ].choices = self.instance.project.offers.not_declined_choices(
            include=self.instance.offer_id
        )

        q = Q(is_active=True)
        if self.instance.pk:
            q |= Q(id__in=self.instance.receivers.values_list("id", flat=True))
        self.fields["receivers"].queryset = User.objects.filter(q)
        self.fields["receivers"].help_text = format_html(
            "{}: {}",
            _("Select team"),
            format_html_join(
                ", ",
                '<a href="#" data-select-receivers="{}">{}</a>',
                [
                    (",".join(str(member.id) for member in team.members.all()), team)
                    for team in Team.objects.prefetch_related("members")
                ],
            ),
        )

    def clean(self):
        data = super().clean()

        if data.get("offer") and data["offer"].is_declined:
            self.add_warning(
                _("The selected offer is declined."), code="offer-is-declined"
            )

        return data

    def save(self):
        if self.instance.pk:
            return super().save()

        instance = super().save(commit=False)
        instance.created_by = self.request.user
        instance.save()
        self.save_m2m()
        return instance


class PlannedWorkSearchForm(Form):
    project = forms.ModelChoiceField(
        queryset=Project.objects.all(),
        required=False,
        widget=Autocomplete(model=Project),
        label="",
    )
    user = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["user"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        if data.get("project"):
            queryset = queryset.filter(project=data.get("project"))
        queryset = self.apply_owned_by(queryset, attribute="user")
        return queryset.select_related(
            "user",
            "project__owned_by",
            "project__customer",
            "project__contact__organization",
        )


@add_prefix("modal")
class PlannedWorkForm(ModelForm):
    user_fields = default_to_current_user = ("user",)

    class Meta:
        model = PlannedWork
        fields = (
            "offer",
            "request",
            "user",
            "title",
            "notes",
            "planned_hours",
        )
        widgets = {
            "notes": Textarea,
        }

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        if not self.project:  # Updating
            self.project = kwargs["instance"].project

        initial = kwargs.setdefault("initial", {})
        request = kwargs["request"]
        for field in ["offer"]:
            if value := request.GET.get(field):
                initial.setdefault(field, value)

        pr = None
        if pr_id := request.POST.get("request") or request.GET.get("request"):
            try:
                pr = self.project.planning_requests.get(pk=pr_id)
            except (PlanningRequest.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.setdefault("offer", pr.offer_id)
                initial.update(
                    {
                        "request": pr.id,
                        "title": pr.title,
                        "planned_hours": pr.requested_hours - pr.planned_hours,
                        "weeks": pr.weeks,
                    }
                )

        if offer_id := request.GET.get("plan_offer"):
            try:
                offer = self.project.offers.get(pk=offer_id)
            except (self.project.offers.model.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "title": offer.title,
                        "notes": offer.description,
                        "planned_hours": offer.services.aggregate(
                            h=Sum("service_hours")
                        )["h"]
                        or Z1,
                    }
                )

        if service_id := request.GET.get("service"):
            try:
                service = self.project.services.get(pk=service_id)
            except (self.project.services.model.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "title": f"{self.project.title}: {service.title}",
                        "notes": service.description,
                        "planned_hours": service.service_hours,
                    }
                )

        super().__init__(*args, **kwargs)
        self.instance.project = self.project

        self.fields[
            "offer"
        ].choices = self.instance.project.offers.not_declined_choices(
            include=self.instance.offer_id
        )
        self.fields["request"].queryset = self.instance.project.planning_requests.all()

        date_from_options = [
            monday(),
            self.instance.weeks and min(self.instance.weeks),
            pr and min(pr.weeks),
        ]
        date_from = min(filter(None, date_from_options)) - dt.timedelta(days=21)

        self.fields["weeks"] = forms.TypedMultipleChoiceField(
            label=capfirst(_("weeks")),
            choices=[
                (
                    day,
                    "KW{} ({} - {})".format(
                        local_date_format(day, fmt="W"),
                        local_date_format(day),
                        local_date_format(day + dt.timedelta(days=6)),
                    ),
                )
                for day in islice(recurring(date_from, "weekly"), 80)
            ],
            widget=forms.SelectMultiple(attrs={"size": 20}),
            initial=self.instance.weeks or [monday()],
            coerce=parse_date,
        )

    def clean(self):
        data = super().clean()

        if data.get("request") and data.get("weeks"):
            outside = [
                week
                for week in data["weeks"]
                if week < data["request"].earliest_start_on
                or week >= data["request"].completion_requested_on
            ]
            if outside:
                self.add_warning(
                    _(
                        "At least one week is outside the requested range of"
                        " %(from)s – %(until)s: %(weeks)s"
                    )
                    % {
                        "from": local_date_format(data["request"].earliest_start_on),
                        "until": local_date_format(
                            data["request"].completion_requested_on
                            - dt.timedelta(days=1)
                        ),
                        "weeks": ", ".join(local_date_format(week) for week in outside),
                    },
                    code="weeks-outside-request",
                )

        if data.get("offer") and data["offer"].is_declined:
            self.add_warning(
                _("The selected offer is declined."), code="offer-is-declined"
            )

        return data

    def save(self):
        instance = super().save(commit=False)
        instance.weeks = self.cleaned_data.get("weeks")
        instance.save()
        return instance



from django.shortcuts import redirect
from django.urls import re_path
from django.utils.translation import gettext_lazy as _

from workbench import generic
from workbench.logbook.forms import (
    BreakForm,
    BreakSearchForm,
    LoggedCostForm,
    LoggedCostMoveForm,
    LoggedCostSearchForm,
    LoggedHoursForm,
    LoggedHoursMoveForm,
    LoggedHoursSearchForm,
)
from workbench.logbook.models import Break, LoggedCost, LoggedHours
from workbench.logbook.views import create


urlpatterns = [
    re_path(
        r"^$", lambda request: redirect("logbook_loggedhours_list"), name="logbook"
    ),
    re_path(
        r"^hours/create/$",
        create,
        {"viewname": "createhours"},
        name="logbook_loggedhours_create",
    ),
    re_path(
        r"^hours/$",
        generic.ListView.as_view(
            model=LoggedHours,
            search_form_class=LoggedHoursSearchForm,
            show_create_button=False,
        ),
        name="logbook_loggedhours_list",
    ),
    re_path(
        r"^hours/(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=LoggedHours),
        name="logbook_loggedhours_detail",
    ),
    re_path(
        r"^hours/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(model=LoggedHours, form_class=LoggedHoursForm),
        name="logbook_loggedhours_update",
    ),
    re_path(
        r"^hours/(?P<pk>\d+)/move/$",
        generic.UpdateView.as_view(
            model=LoggedHours,
            form_class=LoggedHoursMoveForm,
            template_name="modalform.html",
            title=_("Move %(object)s"),
        ),
        name="logbook_loggedhours_move",
    ),
    re_path(
        r"^hours/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=LoggedHours, template_name="modal_confirm_delete.html"
        ),
        name="logbook_loggedhours_delete",
    ),
    re_path(
        r"^costs/create/$",
        create,
        {"viewname": "createcost"},
        name="logbook_loggedcost_create",
    ),
    re_path(
        r"^costs/$",
        generic.ListView.as_view(
            model=LoggedCost,
            search_form_class=LoggedCostSearchForm,
            show_create_button=False,
        ),
        name="logbook_loggedcost_list",
    ),
    re_path(
        r"^costs/(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=LoggedCost),
        name="logbook_loggedcost_detail",
    ),
    re_path(
        r"^costs/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(model=LoggedCost, form_class=LoggedCostForm),
        name="logbook_loggedcost_update",
    ),
    re_path(
        r"^hours/(?P<pk>\d+)/move/$",
        generic.UpdateView.as_view(
            model=LoggedCost,
            form_class=LoggedCostMoveForm,
            template_name="modalform.html",
            title=_("Move %(object)s"),
        ),
        name="logbook_loggedcost_move",
    ),
    re_path(
        r"^costs/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=LoggedCost, template_name="modal_confirm_delete.html"
        ),
        name="logbook_loggedcost_delete",
    ),
    re_path(
        r"^breaks/$",
        generic.ListView.as_view(
            model=Break,
            search_form_class=BreakSearchForm,
            # show_create_button=False,
        ),
        name="logbook_break_list",
    ),
    re_path(
        r"^breaks/(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=Break),
        name="logbook_break_detail",
    ),
    re_path(
        r"^breaks/create/$",
        generic.CreateView.as_view(
            model=Break, form_class=BreakForm, template_name="modalform.html"
        ),
        name="logbook_break_create",
    ),
    re_path(
        r"^breaks/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(
            model=Break, form_class=BreakForm, template_name="modalform.html"
        ),
        name="logbook_break_update",
    ),
    re_path(
        r"^breaks/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=Break, template_name="modal_confirm_delete.html"
        ),
        name="logbook_break_delete",
    ),
]



# Generated by Django 2.2.4 on 2019-08-14 14:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("invoices", "0013_auto_20190630_2220")]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="show_service_details",
            field=models.BooleanField(
                default=True, verbose_name="show service details"
            ),
        ),
        migrations.AddField(
            model_name="recurringinvoice",
            name="show_service_details",
            field=models.BooleanField(
                default=True, verbose_name="show service details"
            ),
        ),
    ]



import datetime as dt
from functools import total_ordering

from django.conf import settings
from django.contrib import messages
from django.db import models
from django.db.models import Q
from django.db.models.expressions import RawSQL
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.projects.models import Project, Service
from workbench.tools.formats import Z2, local_date_format
from workbench.tools.models import ModelWithTotal, SearchQuerySet
from workbench.tools.urls import model_urls
from workbench.tools.validation import raise_if_errors


class OfferQuerySet(SearchQuerySet):
    def in_preparation_choices(self, *, include=None):
        offers = {True: [], False: []}
        for offer in self.filter(Q(status=Offer.IN_PREPARATION) | Q(pk=include)):
            offers[offer.id == include].append((offer.pk, offer))
        return (
            [("", "----------")] + offers[True] + [(_("In preparation"), offers[False])]
        )

    def not_declined_choices(self, *, include=None):
        offers = {True: [], False: []}
        for offer in self.filter(
            ~Q(status=Offer.DECLINED) | Q(pk=include)
        ).select_related("project", "owned_by"):
            offers[offer.is_declined].append((offer.pk, offer))
        return (
            [("", "----------")]
            + offers[False]
            + ([(_("Declined"), offers[True])] if offers[True] else [])
        )

    def maybe_actionable(self, *, user):
        return self.filter(
            Q(status__lt=Offer.ACCEPTED),
            Q(owned_by=user) | Q(owned_by__is_active=False),
        ).select_related("project", "owned_by")

    def offered(self):
        return self.filter(status=Offer.OFFERED)

    def accepted(self):
        return self.filter(status=Offer.ACCEPTED)


@model_urls
@total_ordering
class Offer(ModelWithTotal):
    IN_PREPARATION = 10
    OFFERED = 20
    ACCEPTED = 30
    DECLINED = 40

    STATUS_CHOICES = (
        (IN_PREPARATION, _("In preparation")),
        (OFFERED, _("Offered")),
        (ACCEPTED, _("Accepted")),
        (DECLINED, _("Declined")),
    )

    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    project = models.ForeignKey(
        Project,
        on_delete=models.PROTECT,
        verbose_name=_("project"),
        related_name="offers",
    )

    offered_on = models.DateField(_("offered on"), blank=True, null=True)
    valid_until = models.DateField(_("valid until"), blank=True, null=True)
    closed_on = models.DateField(_("closed on"), blank=True, null=True)

    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"), blank=True)
    owned_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("responsible")
    )

    status = models.PositiveIntegerField(
        _("status"), choices=STATUS_CHOICES, default=IN_PREPARATION
    )

    postal_address = models.TextField(_("postal address"))
    _code = models.IntegerField(_("code"))
    _fts = models.TextField(editable=False, blank=True)

    objects = OfferQuerySet.as_manager()

    class Meta:
        ordering = ["-pk"]
        verbose_name = _("offer")
        verbose_name_plural = _("offers")

    def __str__(self):
        return "{} {} - {}".format(
            self.code, self.title, self.owned_by.get_short_name()
        )

    def __html__(self):
        return format_html(
            "<small>{}</small> {} - {}",
            self.code,
            self.title,
            self.owned_by.get_short_name(),
        )

    def __lt__(self, other):
        # Sorting:
        # - Declined offers are at the end
        # - None (not offered yet) is between declined and other offers
        # - Else order by _code
        myself = (self.is_declined, True, self._code)
        if other is None:
            return myself < (False, True, 1e100)
        elif isinstance(other, Offer):
            return myself < (other.is_declined, True, other._code)
        else:
            return 0

    def get_absolute_url(self):
        return "%s#offer%s" % (self.project.get_absolute_url(), self.pk)

    def save(self, *args, **kwargs):
        new = not self.pk
        if new:
            self._code = RawSQL(
                "SELECT COALESCE(MAX(_code), 0) + 1 FROM offers_offer"
                " WHERE project_id = %s",
                (self.project_id,),
            )
            super().save(*args, **kwargs)
            self.refresh_from_db()

        self._fts = " ".join(
            str(part)
            for part in [
                self.code,
                self.project._fts,
            ]
        )

        if new:
            super().save()
        else:
            super().save(*args, **kwargs)

    save.alters_data = True

    @property
    def code(self):
        return "%s-o%02d" % (self.project.code, self._code)

    def _calculate_total(self):
        self.subtotal = sum(
            (
                service.service_cost
                for service in self.services.all()
                if not service.is_optional
            ),
            Z2,
        )
        super()._calculate_total()

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        errors = {}

        if self.status in (self.OFFERED, self.ACCEPTED, self.DECLINED):
            if not self.offered_on:
                errors.setdefault("offered_on", []).append(
                    _("Offered on date missing for selected state.")
                )
            if not self.valid_until:
                errors.setdefault("valid_until", []).append(
                    _("Valid until date missing for selected state.")
                )

        if self.offered_on and self.valid_until:
            if self.offered_on > self.valid_until:
                errors["valid_until"] = _(
                    "Valid until date has to be after offered on date."
                )

        if self.status >= self.ACCEPTED and not self.closed_on:
            self.closed_on = dt.date.today()
        elif self.status < self.ACCEPTED and self.closed_on:
            self.closed_on = None

        raise_if_errors(errors)

    @property
    def pretty_status(self):
        if self.status == self.IN_PREPARATION:
            return _("In preparation since %(created_at)s") % {
                "created_at": local_date_format(self.created_at.date())
            }
        elif self.status == self.OFFERED:
            if self.project.closed_on:
                return _(
                    "Offered on %(offered_on)s, but project closed on %(closed_on)s"
                ) % {
                    "offered_on": local_date_format(self.offered_on),
                    "closed_on": local_date_format(self.project.closed_on),
                }

            if self.valid_until < dt.date.today():
                return _("Offered on %(offered_on)s, but not valid anymore") % {
                    "offered_on": local_date_format(self.offered_on),
                }

            return _("Offered on %(offered_on)s") % {
                "offered_on": local_date_format(self.offered_on)
            }
        elif self.status in (self.ACCEPTED, self.DECLINED):
            return _("%(status)s on %(closed_on)s") % {
                "status": self.get_status_display(),
                "closed_on": local_date_format(self.closed_on),
            }
        return self.get_status_display()

    @property
    def status_badge(self):
        if self.status == self.OFFERED and self.project.closed_on:
            css = "warning"
        elif self.status == self.OFFERED and self.valid_until < dt.date.today():
            css = "warning"
        else:
            css = {
                self.IN_PREPARATION: "info",
                self.OFFERED: "success",
                self.ACCEPTED: "default",
                self.DECLINED: "danger",
            }[self.status]

        return format_html(
            '<span class="badge badge-{}">{}</span>', css, self.pretty_status
        )

    @property
    def total_title(self):
        return (
            _("Total %(currency)s incl. tax")
            if self.liable_to_vat
            else _("Total %(currency)s")
        ) % {"currency": settings.WORKBENCH.CURRENCY}

    @classmethod
    def allow_create(cls, request):
        if request.path == reverse("offers_offer_create"):
            messages.error(
                request,
                _(
                    "Offers can only be created from projects. Go to the project"
                    " and add services first, then you'll be able to create the"
                    " offer itself."
                ),
            )
            return False
        return True

    @property
    def is_accepted(self):
        return self.status == self.ACCEPTED

    @property
    def is_declined(self):
        return self.status == self.DECLINED

    def copy_to(self, *, project, owned_by):
        offer = Offer.objects.create(
            project=project,
            title=self.title,
            description=self.description,
            owned_by=owned_by,
            discount=self.discount,
            liable_to_vat=self.liable_to_vat,
            show_service_details=self.show_service_details,
        )
        for service in self.services.all():
            new = Service(
                project=project,
                offer=offer,
                allow_logging=service.allow_logging,
                is_optional=service.is_optional,
                role=service.role,
                title=service.title,
                description=service.description,
                position=service.position,
                effort_type=service.effort_type,
                effort_hours=service.effort_hours,
                effort_rate=service.effort_rate,
                cost=service.cost,
                third_party_costs=service.third_party_costs,
            )
            new.save(skip_related_model=True)
        offer.save()
        return offer



# Generated by Django 2.2.3 on 2019-07-12 11:24

import datetime

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("logbook", "0009_auto_20190628_1659")]

    operations = [
        migrations.AlterField(
            model_name="loggedcost",
            name="rendered_on",
            field=models.DateField(
                db_index=True, default=datetime.date.today, verbose_name="rendered on"
            ),
        ),
        migrations.AlterField(
            model_name="loggedhours",
            name="rendered_on",
            field=models.DateField(
                db_index=True, default=datetime.date.today, verbose_name="rendered on"
            ),
        ),
    ]



from collections import defaultdict
from itertools import chain

from django.utils.text import capfirst, slugify
from django.utils.translation import gettext as _

from xlsxdocument import XLSXDocument

from workbench.contacts.models import PostalAddress
from workbench.templatetags.workbench import label
from workbench.tools.formats import Z1


class WorkbenchXLSXDocument(XLSXDocument):
    def logged_hours(self, queryset):
        queryset = queryset.select_related(
            "created_by",
            "rendered_by",
            "service__project__owned_by",
            "invoice_service__invoice__owned_by",
            "invoice_service__invoice__project",
        )

        self.table_from_queryset(
            queryset,
            additional=[
                (capfirst(_("hourly rate")), lambda hours: hours.service.effort_rate),
                (capfirst(_("project")), lambda hours: hours.service.project),
                (
                    capfirst(_("invoice")),
                    lambda hours: hours.invoice_service.invoice
                    if hours.invoice_service
                    else None,
                ),
            ],
        )

        by_service_and_user = defaultdict(lambda: defaultdict(lambda: Z1))
        by_user = defaultdict(lambda: Z1)
        by_service_and_month = defaultdict(lambda: defaultdict(lambda: Z1))
        by_user_and_month = defaultdict(lambda: defaultdict(lambda: Z1))
        by_month = defaultdict(lambda: Z1)

        for h in queryset:
            by_service_and_user[h.service][h.rendered_by] += h.hours
            by_user[h.rendered_by] += h.hours

            month = h.rendered_on.replace(day=1)
            by_service_and_month[h.service][month] += h.hours
            by_user_and_month[h.rendered_by][month] += h.hours
            by_month[month] += h.hours

        self.add_sheet(slugify(_("By service and user")))
        users = sorted(
            set(
                chain.from_iterable(
                    users.keys() for users in by_service_and_user.values()
                )
            )
        )

        self.table(
            [
                capfirst(t)
                for t in [
                    _("project"),
                    _("service"),
                    _("hourly rate"),
                    _("total"),
                    _("total"),
                ]
            ]
            + [user.get_short_name() for user in users],
            [
                [capfirst(_("total")), "", "", sum(by_user.values(), Z1), ""]
                + [by_user.get(user) for user in users]
            ]
            + [
                [
                    service.project,
                    service,
                    service.effort_rate,
                    sum(by_users.values(), Z1),
                    None
                    if service.effort_rate is None
                    else sum(by_users.values(), Z1) * service.effort_rate,
                ]
                + [by_users.get(user) for user in users]
                for service, by_users in sorted(
                    by_service_and_user.items(),
                    key=lambda row: (row[0].project_id, row[0].position),
                )
            ],
        )

        self.add_sheet(slugify(_("By service and month")))
        months = sorted(
            set(
                chain.from_iterable(
                    months.keys() for months in by_service_and_month.values()
                )
            )
        )
        self.table(
            [
                capfirst(t)
                for t in [
                    _("project"),
                    _("service"),
                    _("hourly rate"),
                    _("total"),
                    _("total"),
                ]
            ]
            + months,
            [
                [capfirst(_("total")), "", "", sum(by_month.values(), Z1), ""]
                + [by_month.get(month) for month in months]
            ]
            + [
                [
                    service.project,
                    service,
                    service.effort_rate,
                    sum(by_months.values(), Z1),
                    None
                    if service.effort_rate is None
                    else sum(by_months.values(), Z1) * service.effort_rate,
                ]
                + [by_months.get(month) for month in months]
                for service, by_months in sorted(
                    by_service_and_month.items(),
                    key=lambda row: (row[0].project_id, row[0].position),
                )
            ],
        )

        self.add_sheet(slugify(_("By user and month")))
        months = sorted(
            set(
                chain.from_iterable(
                    months.keys() for months in by_user_and_month.values()
                )
            )
        )
        self.table(
            [
                capfirst(t)
                for t in [
                    _("user"),
                    _("total"),
                ]
            ]
            + months,
            [
                [capfirst(_("total")), sum(by_month.values(), Z1)]
                + [by_month.get(month) for month in months]
            ]
            + [
                [
                    user,
                    sum(by_months.values(), Z1),
                ]
                + [by_months.get(month) for month in months]
                for user, by_months in sorted(
                    by_user_and_month.items(),
                    # key=lambda row: (row[0].project_id, row[0].position),
                )
            ],
        )

    def logged_costs(self, queryset):
        self.table_from_queryset(
            queryset.select_related(
                "created_by",
                "rendered_by",
                "service__project__owned_by",
                "invoice_service__invoice__owned_by",
                "invoice_service__invoice__project",
            ),
            additional=[
                (capfirst(_("service")), lambda cost: cost.service),
                (capfirst(_("project")), lambda cost: cost.service.project),
                (
                    capfirst(_("invoice")),
                    lambda cost: cost.invoice_service.invoice
                    if cost.invoice_service
                    else None,
                ),
            ],
        )

    def people(self, queryset):
        addresses = {
            address.person_id: address
            for address in PostalAddress.objects.filter(person__in=queryset).reverse()
        }

        def getter(field):
            def attribute(person):
                return getattr(addresses.get(person.id), field, "")

            return attribute

        self.table_from_queryset(
            queryset.select_related("organization", "primary_contact"),
            additional=[
                (str(label(PostalAddress, field)), getter(field))
                for field in [
                    "street",
                    "house_number",
                    "address_suffix",
                    "postal_code",
                    "city",
                    "country",
                    "postal_address_override",
                ]
            ],
        )

    def project_budget_statistics(self, statistics):
        self.add_sheet(_("projects"))
        self.table(
            [
                capfirst(_("project")),
                capfirst(_("responsible")),
                _("Offered"),
                _("Logbook"),
                _("Undefined rate"),
                capfirst(_("third party costs")),
                _("Invoiced"),
                _("Not archived"),
                _("Total hours"),
                _("Delta"),
            ],
            [
                (
                    project["project"],
                    project["project"].owned_by.get_short_name(),
                    project["offered"],
                    project["logbook"],
                    project["effort_hours_with_rate_undefined"],
                    project["third_party_costs"],
                    project["invoiced"],
                    project["not_archived"],
                    project["hours"],
                    project["delta"],
                )
                for project in statistics["statistics"]
            ],
        )



# Generated by Django 3.0.2 on 2020-01-14 06:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("accounts", "0007_team"),
    ]

    operations = [
        migrations.RunSQL(
            "SELECT audit_audit_table('accounts_user', ARRAY['password', 'last_login']);"
        ),
        migrations.RunSQL(
            "UPDATE audit_logged_actions SET"
            " row_data=delete(delete(row_data, 'password'), 'last_login'),"
            " changed_fields=delete(delete(changed_fields, 'password'), 'last_login')"
            " WHERE table_name='accounts_user';"
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-16 09:05

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("offers", "0005_offer_show_service_details")]

    operations = [
        migrations.AddField(
            model_name="offer",
            name="total_excl_tax",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="total excl. tax",
            ),
        )
    ]



# Generated by Django 3.1.2 on 2020-11-02 08:00

from django.db import migrations, models

from workbench.tools import search


def forwards(apps, schema_editor):
    Model = apps.get_model("offers.Offer")
    for instance in Model.objects.select_related("project"):
        instance._fts = " ".join(
            str(part)
            for part in [
                "%s-%04d-o%02d"
                % (
                    instance.project.created_at.year,
                    instance.project._code,
                    instance._code,
                ),
                instance.project._fts,
            ]
        )
        instance.save()


class Migration(migrations.Migration):

    dependencies = [
        ("offers", "0009_offer_valid_until"),
    ]

    operations = [
        migrations.AddField(
            model_name="offer",
            name="_fts",
            field=models.TextField(blank=True, editable=False),
        ),
        migrations.RunSQL(
            search.fts(
                "offers_offer", ["title", "description", "postal_address", "_fts"]
            )
        ),
        migrations.RunPython(forwards),
    ]



from django.conf import settings
from django.contrib.admin import (  # noqa
    HORIZONTAL,
    VERTICAL,
    ModelAdmin as ReadWriteModelAdmin,
    RelatedOnlyFieldListFilter,
    StackedInline,
    TabularInline,
    register,
)


class ModelAdmin(ReadWriteModelAdmin):
    def get_actions(self, request):
        return []

    def has_add_permission(self, request, obj=None):
        return settings.WORKBENCH.READ_WRITE_ADMIN

    def has_change_permission(self, request, obj=None):
        return settings.WORKBENCH.READ_WRITE_ADMIN

    def has_delete_permission(self, request, obj=None):
        return settings.WORKBENCH.READ_WRITE_ADMIN



# Generated by Django 2.1.7 on 2019-03-04 21:39

import datetime
from decimal import Decimal

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("invoices", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="LoggedCost",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "rendered_on",
                    models.DateField(
                        default=datetime.date.today, verbose_name="rendered on"
                    ),
                ),
                (
                    "cost",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="cost",
                    ),
                ),
                (
                    "third_party_costs",
                    models.DecimalField(
                        blank=True,
                        decimal_places=2,
                        help_text="Total incl. tax for third-party services.",
                        max_digits=10,
                        null=True,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="third party costs",
                    ),
                ),
                ("description", models.TextField(verbose_name="description")),
                (
                    "archived_at",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="archived at"
                    ),
                ),
            ],
            options={
                "verbose_name": "logged cost",
                "verbose_name_plural": "logged costs",
                "ordering": ("-rendered_on", "-created_at"),
            },
        ),
        migrations.CreateModel(
            name="LoggedHours",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "rendered_on",
                    models.DateField(
                        default=datetime.date.today, verbose_name="rendered on"
                    ),
                ),
                (
                    "hours",
                    models.DecimalField(
                        decimal_places=1,
                        max_digits=4,
                        validators=[
                            django.core.validators.MinValueValidator(Decimal("0.1"))
                        ],
                        verbose_name="hours",
                    ),
                ),
                ("description", models.TextField(verbose_name="description")),
                (
                    "archived_at",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="archived at"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created by",
                    ),
                ),
                (
                    "invoice",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="invoices.Invoice",
                        verbose_name="invoice",
                    ),
                ),
                (
                    "rendered_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="loggedhours",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="rendered by",
                    ),
                ),
            ],
            options={
                "verbose_name": "logged hours",
                "verbose_name_plural": "logged hours",
                "ordering": ("-rendered_on", "-created_at"),
            },
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-22 10:11

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Circle",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
            ],
            options={
                "verbose_name": "circle",
                "verbose_name_plural": "circles",
                "ordering": ["name"],
            },
        ),
        migrations.CreateModel(
            name="Role",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
                (
                    "is_removed",
                    models.BooleanField(default=False, verbose_name="is removed"),
                ),
                (
                    "circle",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="roles",
                        to="circles.Circle",
                        verbose_name="circle",
                    ),
                ),
            ],
            options={
                "verbose_name": "role",
                "verbose_name_plural": "roles",
                "ordering": ["is_removed", "name"],
            },
        ),
    ]



# Generated by Django 3.0.3 on 2020-02-19 11:34

from django.conf import settings
from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [
        ("deals", "0001_initial"),
    ]

    operations = [
        migrations.RunSQL(search.drop_old_shit("deals_deal")),
        migrations.RunSQL(search.create_structure("deals_deal")),
        migrations.RunSQL(
            search.fts(
                "deals_deal",
                ["title", "description", "closing_notice", "_fts"],
            )
        ),
        migrations.RunSQL(
            "SELECT audit_audit_table('deals_valuetype');"
            "SELECT audit_audit_table('deals_value');"
            "SELECT audit_audit_table('deals_deal', ARRAY['_fts', 'fts_document']);"
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-05 10:26

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("deals", "0002_auto_20200219_1234"),
    ]

    operations = [
        migrations.AddField(
            model_name="valuetype",
            name="weekly_target",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="weekly target",
            ),
        ),
    ]



from django import forms
from django.contrib import messages
from django.db.models import Q
from django.forms.models import inlineformset_factory
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.contacts.forms import PostalAddressSelectionForm
from workbench.contacts.models import Organization
from workbench.offers.models import Offer
from workbench.projects.models import Project, Service
from workbench.tools.formats import local_date_format
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea, add_prefix
from workbench.tools.models import ProtectedError, SlowCollector


class OfferSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=(
            ("all", _("All states")),
            ("", _("Open")),
            (_("Exact"), Offer.STATUS_CHOICES),
        ),
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    org = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("s") == "all":
            pass
        elif data.get("s"):
            queryset = queryset.filter(status=data.get("s"))
        else:
            queryset = queryset.filter(status__lte=Offer.OFFERED)
        queryset = self.apply_renamed(queryset, "org", "project__customer")
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related(
            "owned_by",
            "project__owned_by",
            "project__customer",
            "project__contact__organization",
        )


def warn_if_not_in_preparation(form):
    if (
        form.instance.pk
        and form.instance.status >= form.instance.OFFERED
        and form.request.method == "GET"
    ):
        messages.warning(
            form.request,
            _(
                "This offer is not in preparation anymore."
                " I assume you know what you're doing and wont stand in the way."
            ),
        )


class OfferForm(PostalAddressSelectionForm):
    user_fields = default_to_current_user = ("owned_by",)

    class Meta:
        model = Offer
        fields = (
            "offered_on",
            "valid_until",
            "title",
            "description",
            "owned_by",
            "status",
            "postal_address",
            "subtotal",
            "discount",
            "liable_to_vat",
            "show_service_details",
        )
        widgets = {
            "description": Textarea,
            "status": forms.RadioSelect,
            "postal_address": Textarea,
        }

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        if self.project:  # Creating a new offer
            kwargs.setdefault("initial", {}).update({"title": self.project.title})
        else:
            self.project = kwargs["instance"].project

        super().__init__(*args, **kwargs)
        warn_if_not_in_preparation(self)

        self.instance.project = self.project
        self.service_candidates = Service.objects.filter(
            Q(project=self.project), Q(offer__isnull=True) | Q(offer=self.instance)
        )

        self.fields["services"] = forms.ModelMultipleChoiceField(
            queryset=self.service_candidates,
            label=_("services"),
            widget=forms.CheckboxSelectMultiple,
            required=False,
            initial=(
                self.instance.services.values_list("pk", flat=True)
                if self.instance.pk
                else self.service_candidates.values_list("pk", flat=True)
            ),
        )

        self.order_fields(
            field
            for field in list(self.fields)
            if field
            not in {"subtotal", "discount", "liable_to_vat", "show_service_details"}
        )

        self.add_postal_address_selection_if_empty(
            person=self.project.contact, organization=self.project.customer
        )
        self.fields["subtotal"].disabled = True

    def clean(self):
        data = super().clean()
        s_dict = dict(Offer.STATUS_CHOICES)

        if self.instance.closed_on and data["status"] < Offer.ACCEPTED:
            self.add_warning(
                _(
                    "You are attempting to set status to '%(to)s',"
                    " but the offer has already been closed on %(closed)s."
                    " Are you sure?"
                )
                % {
                    "to": s_dict[data["status"]],
                    "closed": local_date_format(self.instance.closed_on),
                },
                code="status-change-but-already-closed",
            )

        if data["status"] == Offer.DECLINED:
            self.add_warning(
                _(
                    "You are setting the offer status to 'Declined'."
                    " However, if you just want to change a few things"
                    " and send the offer to the client again then you"
                    " could just as well put the offer back into preparation."
                ),
                code="yes-please-decline",
            )

        return data

    def save(self):
        instance = super().save(commit=False)
        if not instance.pk:
            instance.save()
        self.cleaned_data["services"].update(offer=instance)
        self.service_candidates.exclude(id__in=self.cleaned_data["services"]).update(
            offer=None
        )
        instance.save()
        instance.project.solely_declined_offers_warning(request=self.request)
        return instance


class OfferPricingForm(ModelForm):
    class Meta:
        model = Offer
        fields = ("subtotal", "discount", "total_excl_tax", "liable_to_vat", "total")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn_if_not_in_preparation(self)

        self.fields["subtotal"].disabled = True
        self.fields["total_excl_tax"].disabled = True
        self.fields["total"].disabled = True

        kwargs.pop("request")
        self.formsets = (
            {"services": ServiceFormset(*args, **kwargs)} if self.instance.pk else {}
        )

    def is_valid(self):
        return all(
            [super().is_valid()]
            + [formset.is_valid() for formset in self.formsets.values()]
        )

    def save(self, commit=True):
        for formset in self.formsets.values():
            formset.save()
        return super().save()


class ServiceForm(forms.ModelForm):
    class Meta:
        model = Service
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and self.instance.project.flat_rate is not None:
            self.fields["effort_type"].disabled = True
            self.fields["effort_rate"].disabled = True

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.save(skip_related_model=True)
        return instance


ServiceFormset = inlineformset_factory(
    Offer,
    Service,
    form=ServiceForm,
    fields=("title", "effort_type", "effort_rate", "effort_hours", "cost"),
    extra=0,
    can_delete=False,
)


@add_prefix("modal")
class OfferCopyForm(Form):
    project = forms.ModelChoiceField(
        queryset=Project.objects.all(),
        widget=Autocomplete(model=Project, params={"only_open": "on"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project")
        super().__init__(*args, **kwargs)

    def clean_project(self):
        project = self.cleaned_data.get("project")
        if project == self.project:
            self.add_warning(
                _("Same project selected as offer already belongs to."),
                code="same-project",
            )
        return project


class OfferDeleteForm(ModelForm):
    class Meta:
        model = Offer
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        collector = SlowCollector(using=self.instance._state.db)
        try:
            collector.collect(self.instance.services.all())
        except ProtectedError:
            pass
        else:
            self.fields["delete_services"] = forms.BooleanField(
                label=_("Delete offers' services?"), required=False
            )

    def delete(self):
        if self.cleaned_data.get("delete_services"):
            self.instance.services.all().delete()
        self.instance.delete()


@add_prefix("modal")
class OfferAutocompleteForm(forms.Form):
    offer = forms.ModelChoiceField(
        queryset=Offer.objects.all(),
        widget=Autocomplete(model=Offer),
        label="",
    )



import datetime as dt

from django.utils.translation import gettext as _

from workbench.tools.validation import monday


def date_ranges():
    this_month = dt.date.today().replace(day=1)
    last_month = (this_month - dt.timedelta(days=1)).replace(day=1)
    next_month = (this_month + dt.timedelta(days=31)).replace(day=1)

    this_quarter = dt.date(this_month.year, 1 + (this_month.month - 1) // 3 * 3, 1)
    last_quarter = (this_quarter - dt.timedelta(days=75)).replace(day=1)
    next_quarter = (this_quarter + dt.timedelta(days=105)).replace(day=1)

    return [
        (
            (monday() + dt.timedelta(days=0)).isoformat(),
            (monday() + dt.timedelta(days=6)).isoformat(),
            _("this week"),
        ),
        (
            (monday() - dt.timedelta(days=7)).isoformat(),
            (monday() - dt.timedelta(days=1)).isoformat(),
            _("last week"),
        ),
        (
            this_month.isoformat(),
            (next_month - dt.timedelta(days=1)).isoformat(),
            _("this month"),
        ),
        (
            last_month.isoformat(),
            (this_month - dt.timedelta(days=1)).isoformat(),
            _("last month"),
        ),
        (
            this_quarter.isoformat(),
            (next_quarter - dt.timedelta(days=1)).isoformat(),
            _("this quarter"),
        ),
        (
            last_quarter.isoformat(),
            (this_quarter - dt.timedelta(days=1)).isoformat(),
            _("last quarter"),
        ),
        (
            dt.date(this_month.year, 1, 1).isoformat(),
            dt.date(this_month.year, 12, 31).isoformat(),
            _("this year"),
        ),
        (
            dt.date(this_month.year - 1, 1, 1).isoformat(),
            dt.date(this_month.year - 1, 12, 31).isoformat(),
            _("last year"),
        ),
    ]



import datetime as dt

from django.core import mail
from django.test import TestCase
from django.utils.translation import deactivate_all

from time_machine import travel

from workbench import factories
from workbench.invoices.models import Invoice, RecurringInvoice
from workbench.invoices.tasks import create_recurring_invoices_and_notify
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days


class RecurringTest(TestCase):
    def setUp(self):
        deactivate_all()

    def test_recurring_invoice(self):
        """Creating recurring invoices and invoices from a RI works as expected"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )
        pa = factories.PostalAddressFactory.create(person=person)

        ri = RecurringInvoice.objects.create(
            customer=person.organization,
            contact=person,
            title="Recurring invoice",
            description="",
            owned_by=person.primary_contact,
            postal_address=pa.postal_address,
            starts_on=dt.date(2018, 1, 1),
            ends_on=dt.date(2018, 12, 31),
            periodicity="monthly",
            subtotal=200,
        )

        self.assertEqual(ri.next_period_starts_on, None)

        with travel("2019-01-01 12:00"):
            self.assertEqual(len(ri.create_invoices()), 12)

        ri.ends_on = None
        ri.save()

        with travel("2019-01-01 12:00"):
            self.assertEqual(len(ri.create_invoices()), 1)

        # Continue generating invoices after January '19
        invoices = []
        for ri in RecurringInvoice.objects.renewal_candidates():
            invoices.extend(ri.create_invoices())
        self.assertTrue(len(invoices) > 1)
        self.assertEqual({invoice.type for invoice in invoices}, {Invoice.FIXED})

    def test_creation(self):
        """Creation of recurring invoices, with the customer/contaact pre-form"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )

        self.client.force_login(person.primary_contact)

        response = self.client.get("/recurring-invoices/create/")
        self.assertNotContains(response, 'name="title"')

        response = self.client.get(
            "/recurring-invoices/create/?contact={}".format(person.pk)
        )
        self.assertContains(
            response,
            # No value!
            '<input type="number" name="third_party_costs" step="0.01" class="form-control" required id="id_third_party_costs">',  # noqa
            html=True,
        )
        # print(response, response.content.decode("utf-8"))

        response = self.client.post(
            "/recurring-invoices/create/?contact={}".format(person.pk),
            {
                "customer": person.organization_id,
                # "contact": person.id,
                "title": "recur",
                "owned_by": person.primary_contact_id,
                "starts_on": dt.date.today().isoformat(),
                "periodicity": "yearly",
                "create_invoice_on_day": -20,
                "subtotal": 500,
                "discount": 0,
                "liable_to_vat": "on",
                "third_party_costs": 0,
            },
        )
        self.assertContains(response, "No contact selected.")

        response = self.client.post(
            "/recurring-invoices/create/?contact={}".format(person.pk),
            {
                "customer": person.organization_id,
                "contact": person.id,
                "title": "recur",
                "owned_by": person.primary_contact_id,
                "starts_on": dt.date.today().isoformat(),
                "periodicity": "yearly",
                "create_invoice_on_day": -20,
                "subtotal": 500,
                "discount": 0,
                "liable_to_vat": "on",
                "third_party_costs": 0,
                "postal_address": "Anything",
            },
        )
        self.assertContains(response, 'value="short-postal-address"')

        response = self.client.post(
            "/recurring-invoices/create/?contact={}".format(person.pk),
            {
                "customer": person.organization_id,
                "contact": person.id,
                "title": "recur",
                "owned_by": person.primary_contact_id,
                "starts_on": dt.date.today().isoformat(),
                "periodicity": "yearly",
                "create_invoice_on_day": -20,
                "subtotal": 500,
                "discount": 0,
                "liable_to_vat": "on",
                "third_party_costs": 0,
                "postal_address": "Anything\nStreet\nCity",
            },
        )

        ri = RecurringInvoice.objects.get()
        self.assertRedirects(response, ri.urls["detail"])

        factories.PostalAddressFactory.create(person=person)
        response = self.client.get(ri.urls["update"])
        self.assertContains(response, 'name="postal_address"')
        self.assertContains(response, 'data-field-value="')

        response = self.client.get(ri.urls["detail"])
        self.assertContains(response, "?create_invoices=1")

        response = self.client.get(ri.urls["detail"] + "?create_invoices=1")
        self.assertRedirects(response, Invoice.urls["list"])
        self.assertEqual(messages(response), ["Created 1 invoice."])

        response = self.client.get(ri.urls["detail"] + "?create_invoices=1")
        self.assertRedirects(response, ri.urls["detail"])
        self.assertEqual(messages(response), ["Created 0 invoices."])

    def test_create_and_notify(self):
        """The recurring invoices task notifies owners"""
        factories.RecurringInvoiceFactory.create()

        create_recurring_invoices_and_notify()
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "recurring invoices")

        create_recurring_invoices_and_notify()
        self.assertEqual(len(mail.outbox), 1)

    def test_list(self):
        """Filter form smoke test"""
        factories.RecurringInvoiceFactory.create()

        user = factories.UserFactory.create()
        self.client.force_login(user)

        code = check_code(self, "/recurring-invoices/")
        code("")
        code("q=test")
        code("s=all")
        code("s=closed")
        code("org={}".format(factories.OrganizationFactory.create().pk))
        code("owned_by={}".format(user.id))
        code("owned_by=-1")  # mine
        code("owned_by=0")  # only inactive

    def test_pretty_status(self):
        """Recurring invoice status badge"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )

        ri = RecurringInvoice.objects.create(
            customer=person.organization,
            contact=person,
            title="Recurring invoice",
            owned_by=person.primary_contact,
            starts_on=dt.date(2018, 1, 1),
            ends_on=dt.date(2050, 12, 31),
            periodicity="monthly",
            subtotal=200,
        )
        self.assertEqual(ri.pretty_status, "monthly from 01.01.2018 until 31.12.2050")
        self.assertEqual(
            ri.pretty_next_period,
            "Next period starts on 01.01.2018, invoice will be created on 12.12.2017",
        )

        ri = RecurringInvoice.objects.create(
            customer=person.organization,
            contact=person,
            title="Recurring invoice",
            owned_by=person.primary_contact,
            starts_on=dt.date(2018, 1, 1),
            ends_on=None,
            periodicity="monthly",
            subtotal=200,
        )
        self.assertEqual(ri.pretty_status, "monthly from 01.01.2018")
        self.assertEqual(
            ri.pretty_next_period,
            "Next period starts on 01.01.2018, invoice will be created on 12.12.2017",
        )

        ri = RecurringInvoice.objects.create(
            customer=person.organization,
            contact=person,
            title="Recurring invoice",
            owned_by=person.primary_contact,
            starts_on=dt.date(2018, 1, 1),
            ends_on=dt.date(2018, 1, 31),
            next_period_starts_on=dt.date(2018, 2, 1),
            periodicity="monthly",
            subtotal=200,
        )
        self.assertEqual(ri.pretty_status, "monthly from 01.01.2018 until 31.01.2018")
        self.assertEqual(ri.pretty_next_period, "")

    def test_pre_form(self):
        """Pre-form branch coverage"""
        self.client.force_login(factories.UserFactory.create())

        # pre_form does not have these fields
        response = self.client.get(RecurringInvoice.urls["create"])
        self.assertContains(response, 'method="GET"')
        self.assertNotContains(response, 'id="id_title"')
        self.assertNotContains(response, 'id="id_description"')

        # Nonexistant entries
        response = self.client.get(RecurringInvoice.urls["create"] + "?contact=0")
        self.assertContains(response, 'method="GET"')
        self.assertNotContains(response, 'id="id_title"')
        self.assertNotContains(response, 'id="id_description"')

        response = self.client.get(RecurringInvoice.urls["create"] + "?customer=0")
        self.assertContains(response, 'method="GET"')
        self.assertNotContains(response, 'id="id_title"')
        self.assertNotContains(response, 'id="id_description"')

        # Existing
        organization = factories.OrganizationFactory.create()
        response = self.client.get(
            RecurringInvoice.urls["create"] + "?customer={}".format(organization.pk)
        )
        self.assertContains(response, 'method="POST"')

        person = factories.PersonFactory.create()
        response = self.client.get(
            RecurringInvoice.urls["create"] + "?contact={}".format(person.pk)
        )
        self.assertContains(response, 'method="POST"')

    def test_copy(self):
        """Copying recurring invoices"""
        invoice = factories.RecurringInvoiceFactory.create()
        self.client.force_login(invoice.owned_by)

        response = self.client.get(invoice.urls["create"] + "?copy=" + str(invoice.pk))
        self.assertContains(response, 'value="{}"'.format(invoice.title))
        # print(response, response.content.decode("utf-8"))

        response = self.client.get(invoice.urls["create"] + "?copy=blub")
        self.assertEqual(response.status_code, 200)  # No crash

    def test_renewal_candidates(self):
        """Renewal candidates depends on start date and the day of period when
        the invoice is created"""
        r1 = factories.RecurringInvoiceFactory.create(
            starts_on=in_days(10),
            periodicity="monthly",
        )
        r2 = factories.RecurringInvoiceFactory.create(
            starts_on=in_days(30),
            periodicity="monthly",
        )

        self.assertEqual(set(RecurringInvoice.objects.renewal_candidates()), {r1})

        r3 = factories.RecurringInvoiceFactory.create(
            starts_on=in_days(-250),
            periodicity="yearly",
            create_invoice_on_day=300,
        )
        r4 = factories.RecurringInvoiceFactory.create(
            starts_on=in_days(-350),
            periodicity="yearly",
            create_invoice_on_day=300,
        )

        self.assertEqual(set(RecurringInvoice.objects.renewal_candidates()), {r1, r4})

        r2, r3  # Using those variables

    def test_positive_create_invoice_on_day(self):
        """Invoice creation can be moved within the period"""
        r1 = factories.RecurringInvoiceFactory.create(
            starts_on=dt.date.today(),
            periodicity="monthly",
        )
        r2 = factories.RecurringInvoiceFactory.create(
            starts_on=dt.date.today(),
            periodicity="monthly",
            # Wait with invoice creation until 10 days into the period:
            create_invoice_on_day=10,
        )

        self.assertEqual(len(r1.create_invoices()), 1)
        self.assertEqual(len(r2.create_invoices()), 0)

    def test_create_project(self):
        """Recurring invoices may optionally create projects"""
        r = factories.RecurringInvoiceFactory.create(create_project=True)

        invoices = r.create_invoices()
        self.assertIsNotNone(invoices[0].project)
        self.assertEqual(invoices[0].type, invoices[0].DOWN_PAYMENT)



from workbench.tools import admin

from . import models


@admin.register(models.Circle)
class CircleAdmin(admin.ModelAdmin):
    pass


@admin.register(models.Role)
class RoleAdmin(admin.ReadWriteModelAdmin):
    list_display = ["name", "circle", "for_circle", "is_removed", "work_category"]
    list_editable = ["work_category"]
    list_filter = ["work_category", "circle", "is_removed"]
    readonly_fields = ["circle", "name", "for_circle", "is_removed"]

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



import datetime as dt
from functools import total_ordering

from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.contacts.models import Organization, Person
from workbench.tools.formats import Z2, currency, local_date_format
from workbench.tools.models import Model, MoneyField, SearchQuerySet
from workbench.tools.urls import model_urls


class NotArchivedQuerySet(models.QuerySet):
    def active(self, include=None):
        return self.filter(Q(is_archived=False) | Q(id=include))


class AttributeGroup(models.Model):
    title = models.CharField(_("title"), max_length=200)
    position = models.PositiveIntegerField(_("position"), default=0)
    is_archived = models.BooleanField(_("is archived"), default=False)
    is_required = models.BooleanField(_("is required"), default=True)

    class Meta:
        ordering = ("position", "id")
        verbose_name = _("attribute group")
        verbose_name_plural = _("attribute groups")

    def __str__(self):
        return self.title


class Attribute(models.Model):
    group = models.ForeignKey(
        AttributeGroup,
        on_delete=models.CASCADE,
        related_name="attributes",
        verbose_name=_("attribute group"),
    )
    title = models.CharField(_("title"), max_length=200)
    position = models.PositiveIntegerField(_("position"), default=0)
    is_archived = models.BooleanField(_("is archived"), default=False)

    objects = NotArchivedQuerySet.as_manager()

    class Meta:
        ordering = ("position", "id")
        verbose_name = _("attribute")
        verbose_name_plural = _("attributes")

    def __str__(self):
        return self.title


class ClosingType(models.Model):
    title = models.CharField(_("title"), max_length=200)
    represents_a_win = models.BooleanField(_("represents a win"), default=False)
    position = models.PositiveIntegerField(_("position"), default=0)

    class Meta:
        ordering = ("position", "id")
        verbose_name = _("closing type")
        verbose_name_plural = _("closing types")

    def __str__(self):
        return self.title


class DealQuerySet(SearchQuerySet):
    def maybe_actionable(self, *, user):
        return self.filter(
            Q(status=Deal.OPEN), Q(owned_by=user) | Q(owned_by__is_active=False)
        ).select_related("owned_by")


@model_urls
class Deal(Model):
    OPEN = 10
    ACCEPTED = 20
    DECLINED = 30

    STATUS_CHOICES = (
        (OPEN, _("Open")),
        (ACCEPTED, _("Accepted")),
        (DECLINED, _("Declined")),
    )

    UNKNOWN = 10
    NORMAL = 20
    HIGH = 30

    PROBABILITY_CHOICES = [
        (UNKNOWN, _("unknown")),
        (NORMAL, _("normal")),
        (HIGH, _("high")),
    ]

    customer = models.ForeignKey(
        Organization, on_delete=models.PROTECT, verbose_name=_("customer")
    )
    contact = models.ForeignKey(
        Person,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("contact"),
    )

    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"), blank=True)
    owned_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("responsible")
    )
    value = MoneyField(_("value"))

    status = models.PositiveIntegerField(
        _("status"), choices=STATUS_CHOICES, default=OPEN
    )
    probability = models.IntegerField(
        _("probability"), choices=PROBABILITY_CHOICES, default=UNKNOWN
    )
    decision_expected_on = models.DateField(
        _("decision expected on"), blank=True, null=True
    )

    created_at = models.DateTimeField(_("created at"), default=timezone.now)

    attributes = models.ManyToManyField(
        Attribute,
        verbose_name=_("attributes"),
        through="DealAttribute",
    )

    closed_on = models.DateField(_("closed on"), blank=True, null=True)
    closing_type = models.ForeignKey(
        ClosingType,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("closing type"),
    )
    closing_notice = models.TextField(_("closing notice"), blank=True)

    _fts = models.TextField(editable=False, blank=True)

    related_offers = models.ManyToManyField(
        "offers.Offer",
        blank=True,
        related_name="deals",
        verbose_name=_("related offers"),
    )

    objects = DealQuerySet.as_manager()

    class Meta:
        ordering = ["-pk"]
        verbose_name = _("deal")
        verbose_name_plural = _("deals")

    def __str__(self):
        return "%s %s - %s" % (self.code, self.title, self.owned_by.get_short_name())

    def __html__(self):
        return format_html(
            "<small>{}</small> {} - {}",
            self.code,
            self.title,
            self.owned_by.get_short_name(),
        )

    def get_related_offers(self):
        return self.related_offers.select_related("owned_by", "project")

    def save(self, *args, **kwargs):
        skip_value_calculation = kwargs.pop("skip_value_calculation", False)

        if not skip_value_calculation:
            self.value = sum((v.value for v in self.values.all()), Z2)

        self._fts = " ".join(
            str(part)
            for part in [
                self.code,
                self.customer.name,
                self.contact.full_name if self.contact else "",
            ]
        )
        super().save(*args, **kwargs)

    save.alters_data = True

    @property
    def pretty_status(self):
        d = {
            "created_at": local_date_format(self.created_at.date()),
            "closed_on": self.closed_on and local_date_format(self.closed_on),
            "decision_expected_on": self.decision_expected_on
            and local_date_format(self.decision_expected_on),
            "status": self.get_status_display(),
        }

        if self.status != self.OPEN:
            return _("%(status)s on %(closed_on)s") % d
        if self.decision_expected_on:
            return _("Decision expected on %(decision_expected_on)s") % d
        return _("Open since %(created_at)s") % d

    @property
    def status_badge(self):
        if self.status != self.OPEN:
            css = {self.ACCEPTED: "success", self.DECLINED: "danger"}[self.status]
        elif self.decision_expected_on:
            css = "warning" if self.decision_expected_on < dt.date.today() else "info"
        else:
            open_since = (dt.date.today() - self.created_at.date()).days
            if (
                open_since
                > {self.UNKNOWN: 90, self.NORMAL: 45, self.HIGH: 20}[self.probability]
            ):
                css = "caveat"
            else:
                css = "info"

        return format_html(
            '<span class="badge badge-{}">{}</span>', css, self.pretty_status
        )

    @property
    def pretty_closing_type(self):
        return self.closing_type or _("<closing type missing>")


class DealAttribute(models.Model):
    deal = models.ForeignKey(Deal, on_delete=models.CASCADE, verbose_name=_("deal"))
    attribute = models.ForeignKey(
        Attribute, on_delete=models.PROTECT, verbose_name=_("attribute")
    )

    class Meta:
        verbose_name = _("deal attribute")
        verbose_name_plural = _("deal attributes")

    def __str__(self):
        return "{} - {}".format(self.deal, self.attribute)


@total_ordering
class ValueType(models.Model):
    title = models.CharField(_("title"), max_length=200)
    position = models.PositiveIntegerField(_("position"), default=0)
    is_archived = models.BooleanField(_("is archived"), default=False)
    weekly_target = MoneyField(_("weekly target"), blank=True, null=True)

    class Meta:
        ordering = ("position", "id")
        verbose_name = _("value type")
        verbose_name_plural = _("value types")

    def __str__(self):
        return self.title

    def __lt__(self, other):
        return (
            (self.position, -self.pk) < (other.position, -other.pk)
            if isinstance(other, self.__class__)
            else 1
        )


class Value(models.Model):
    deal = models.ForeignKey(
        Deal, on_delete=models.CASCADE, related_name="values", verbose_name=_("deal")
    )
    type = models.ForeignKey(
        ValueType, on_delete=models.PROTECT, verbose_name=_("type")
    )
    value = MoneyField(_("value"))

    class Meta:
        ordering = ["type"]
        unique_together = [("deal", "type")]
        verbose_name = _("value")
        verbose_name_plural = _("values")

    def __str__(self):
        return currency(self.value)



# Generated by Django 2.1.7 on 2019-03-21 13:48

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("logbook", "0005_auto_20190312_1254")]

    operations = [
        migrations.RunSQL(search.drop_old_shit("logbook_loggedcost")),
        migrations.RunSQL(search.drop_old_shit("logbook_loggedhours")),
        migrations.RunSQL(search.create_structure("logbook_loggedcost")),
        migrations.RunSQL(search.create_structure("logbook_loggedhours")),
        migrations.RunSQL(search.fts("logbook_loggedcost", ["description"])),
        migrations.RunSQL(search.fts("logbook_loggedhours", ["description"])),
    ]



from django.core.management import BaseCommand
from django.utils.translation import activate

from workbench.accounts.middleware import set_user_name
from workbench.invoices.tasks import create_recurring_invoices_and_notify
from workbench.reporting.accounting import send_accounting_files
from workbench.reporting.tasks import create_accruals_for_last_month


class Command(BaseCommand):
    help = "Fairy tasks"

    def handle(self, **options):
        activate("de")
        set_user_name("Fairy tasks")
        create_accruals_for_last_month()
        create_recurring_invoices_and_notify()
        send_accounting_files()



import datetime as dt
import re
from functools import wraps

from django.contrib import messages
from django.core.exceptions import ValidationError
from django.shortcuts import redirect
from django.utils.translation import gettext as _


def monday(day=None):
    day = day or dt.date.today()
    return day - dt.timedelta(days=day.weekday())


def logbook_lock():
    day = dt.date.today()
    return max(day.replace(month=1, day=1), day - dt.timedelta(days=day.weekday()))


def raise_if_errors(errors, exclude=None):
    if errors:
        if set(exclude or ()) & errors.keys():
            raise ValidationError(", ".join(str(e) for e in errors.values()))
        raise ValidationError(errors)


def filter_form(form_class):
    def decorator(view):
        @wraps(view)
        def inner(request, *args, **kwargs):
            form = form_class(request.GET, request=request)
            if not form.is_valid():
                messages.warning(request, _("Form was invalid."))
                return redirect(".")
            return view(request, form, *args, **kwargs)

        return inner

    return decorator


def in_days(days):
    return dt.date.today() + dt.timedelta(days=days)


UNSPECIFIC_TITLE = re.compile(r"(general|allg|unspec)", re.I)
WHITESPACE = re.compile(r"\s+")


def is_title_specific(title):
    return not UNSPECIFIC_TITLE.search(title) or len(WHITESPACE.split(title)) >= 5



# Generated by Django 2.1.7 on 2019-03-12 14:35

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("services", "0002_auto_20190304_2247")]

    operations = [
        migrations.RenameField(
            model_name="servicetype",
            old_name="billing_per_hour",
            new_name="hourly_rate",
        ),
        migrations.AlterField(
            model_name="servicetype",
            name="hourly_rate",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="hourly rate",
            ),
        ),
    ]



# Generated by Django 2.2 on 2019-04-12 15:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("projects", "0008_service_is_logging_prohibited")]

    operations = [
        migrations.AddField(
            model_name="service",
            name="is_optional",
            field=models.BooleanField(
                default=False,
                help_text="Optional services to not count towards the offer total.",
                verbose_name="is optional",
            ),
        )
    ]



# Generated by Django 2.1.7 on 2019-03-08 16:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("credit_control", "0001_initial")]

    operations = [
        migrations.RunSQL("SELECT audit_audit_table('credit_control_creditentry');", "")
    ]



from workbench.credit_control import models
from workbench.tools import admin


@admin.register(models.Ledger)
class LedgerAdmin(admin.ReadWriteModelAdmin):
    radio_fields = {"parser": admin.VERTICAL}


@admin.register(models.CreditEntry)
class CreditEntryAdmin(admin.ModelAdmin):
    list_display = [
        "reference_number",
        "value_date",
        "total",
        "payment_notice",
        "invoice",
        "notes",
    ]
    list_select_related = ["invoice__owned_by", "invoice__project"]
    raw_id_fields = ["invoice"]



import datetime as dt
from itertools import islice

from django.test import TestCase

from workbench.invoices.utils import next_valid_day, recurring


class UtilsTest(TestCase):
    def test_next_valid_day(self):
        """The next_valid_day returns valid days :-)"""
        self.assertEqual(next_valid_day(2016, 2, 28), dt.date(2016, 2, 28))
        self.assertEqual(next_valid_day(2016, 2, 29), dt.date(2016, 2, 29))

        self.assertEqual(next_valid_day(2017, 2, 28), dt.date(2017, 2, 28))
        self.assertEqual(next_valid_day(2017, 2, 29), dt.date(2017, 3, 1))

        # Days>31 may not increment months by more than 1
        self.assertEqual(next_valid_day(2018, 1, 65), dt.date(2018, 2, 1))

        # Months>12 may increment years by more than 1
        self.assertEqual(next_valid_day(2018, 27, 1), dt.date(2020, 3, 1))

    def test_recurring(self):
        """The recurring() utilty returns expected values"""
        self.assertEqual(
            list(islice(recurring(dt.date(2016, 2, 29), "yearly"), 5)),
            [
                dt.date(2016, 2, 29),
                dt.date(2017, 3, 1),
                dt.date(2018, 3, 1),
                dt.date(2019, 3, 1),
                dt.date(2020, 2, 29),
            ],
        )

        self.assertEqual(
            list(islice(recurring(dt.date(2016, 1, 31), "quarterly"), 5)),
            [
                dt.date(2016, 1, 31),
                dt.date(2016, 5, 1),
                dt.date(2016, 7, 31),
                dt.date(2016, 10, 31),
                dt.date(2017, 1, 31),
            ],
        )

        self.assertEqual(
            list(islice(recurring(dt.date(2016, 1, 31), "monthly"), 5)),
            [
                dt.date(2016, 1, 31),
                dt.date(2016, 3, 1),
                dt.date(2016, 3, 31),
                dt.date(2016, 5, 1),
                dt.date(2016, 5, 31),
            ],
        )

        self.assertEqual(
            list(islice(recurring(dt.date(2016, 1, 1), "weekly"), 5)),
            [
                dt.date(2016, 1, 1),
                dt.date(2016, 1, 8),
                dt.date(2016, 1, 15),
                dt.date(2016, 1, 22),
                dt.date(2016, 1, 29),
            ],
        )

        with self.assertRaises(ValueError):
            list(islice(recurring(dt.date(2016, 1, 1), "unknown"), 5))



from django.shortcuts import get_object_or_404, redirect
from django.urls import re_path

from workbench import generic
from workbench.accounts.features import controlling_only
from workbench.offers.forms import OfferForm, OfferPricingForm, OfferSearchForm
from workbench.offers.models import Offer
from workbench.offers.views import OfferDeleteView, OfferPDFView, copy_offer


urlpatterns = [
    re_path(
        r"^$",
        controlling_only(
            generic.ListView.as_view(model=Offer, search_form_class=OfferSearchForm)
        ),
        name="offers_offer_list",
    ),
    re_path(
        r"^autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Offer,
            queryset=Offer.objects.select_related("owned_by", "project"),
        ),
        name="offers_offer_autocomplete",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        lambda request, pk: redirect(get_object_or_404(Offer, pk=pk)),
        name="offers_offer_detail",
    ),
    re_path(
        r"^create/$",
        controlling_only(generic.CreateView.as_view(model=Offer, form_class=OfferForm)),
        name="offers_offer_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        controlling_only(generic.UpdateView.as_view(form_class=OfferForm, model=Offer)),
        name="offers_offer_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/pricing/$",
        controlling_only(
            generic.UpdateView.as_view(
                form_class=OfferPricingForm,
                model=Offer,
                template_name_suffix="_pricing",
            )
        ),
        name="offers_offer_pricing",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        controlling_only(OfferDeleteView.as_view(model=Offer)),
        name="offers_offer_delete",
    ),
    re_path(
        r"^(?P<pk>\d+)/pdf/$",
        controlling_only(OfferPDFView.as_view()),
        name="offers_offer_pdf",
    ),
    re_path(
        r"^(?P<pk>\d+)/copy/$", controlling_only(copy_offer), name="offers_offer_copy"
    ),
]



import os
import sys
import types
from collections import defaultdict

from django.utils.translation import gettext_lazy as _

import dj_database_url
import dj_email_url
from speckenv import env


BASE_DIR = os.path.dirname(os.path.dirname(__file__))

SECRET_KEY = env("SECRET_KEY", required=True)
DEBUG = env("DEBUG", default=bool({"runserver"}.intersection(sys.argv)))
TESTING = env("TESTING", default="test" in sys.argv)
LIVE = env("LIVE", default=False)
ALLOWED_HOSTS = env("ALLOWED_HOSTS", default=[])
ADMINS = MANAGERS = [("Matthias Kestenholz", "mk@feinheit.ch")]
DEFAULT_FROM_EMAIL = SERVER_EMAIL = "workbench@workbench.feinheit.ch"
BCC = env("BCC", default=[row[1] for row in MANAGERS])

DEBUG_TOOLBAR = DEBUG and not TESTING

INSTALLED_APPS = [
    a
    for a in [
        "workbench",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.postgres",
        "admin_ordering",
        "bootstrap4",
        "django_countries",
        "authlib",
        "corsheaders",
        "fineforms",
        "webpack_loader",
        "django.forms",
        "workbench.accounts",
        "workbench.audit",
        "workbench.awt",
        "workbench.circles",
        "workbench.contacts",
        "workbench.credit_control",
        "workbench.deals",
        "workbench.expenses",
        "workbench.invoices",
        "workbench.logbook",
        "workbench.notes",
        "workbench.offers",
        "workbench.planning",
        "workbench.projects",
        "workbench.reporting",
        "workbench.services",
        "workbench.timer",
        "debug_toolbar" if DEBUG_TOOLBAR else "",
    ]
    if a
]

AUTH_USER_MODEL = "accounts.User"
LOGIN_REDIRECT_URL = "/"
LOGIN_REQUIRED_EXEMPT = (
    "/accounts",
    "/favicon",
    "/robots",
    "/sitemap",
    "/create-timestamp",
    "/list-timestamps",
    "/timestamps-controller",
)

MIDDLEWARE = [
    m
    for m in [
        "django.middleware.security.SecurityMiddleware" if LIVE else "",
        "debug_toolbar.middleware.DebugToolbarMiddleware" if DEBUG_TOOLBAR else "",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.locale.LocaleMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
        "workbench.accounts.middleware.user_middleware",
        "workbench.middleware.history_fallback",
    ]
    if m
]

ROOT_URLCONF = "workbench.urls"
WSGI_APPLICATION = "wsgi.application"
WEBPACK_LOADER = {
    "DEFAULT": {
        "STATS_FILE": os.path.join(
            BASE_DIR, "static", "webpack-stats-%s.json" % ("dev" if DEBUG else "prod")
        )
    }
}

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "workbench", "templates")],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "workbench.context_processors.workbench",
            ],
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                "django.template.loaders.app_directories.Loader",
            ]
            if DEBUG
            else [
                (
                    "django.template.loaders.cached.Loader",
                    [
                        "django.template.loaders.filesystem.Loader",
                        "django.template.loaders.app_directories.Loader",
                    ],
                )
            ],
            "debug": DEBUG,
        },
    }
]

LOCALE_PATHS = [os.path.join(BASE_DIR, "conf", "locale")]

AUTHENTICATION_BACKENDS = ["authlib.backends.EmailBackend"]

DATABASES = {"default": dj_database_url.config(default="sqlite:///db.sqlite3")}
ATOMIC_REQUESTS = True

LANGUAGE_CODE = "de"
LANGUAGES = [("en", _("english")), ("de", _("german"))]
TIME_ZONE = "Europe/Zurich"
USE_I18N = True
USE_L10N = False
USE_TZ = True

if LIVE:  # pragma: no cover
    STATICFILES_STORAGE = (
        "django.contrib.staticfiles.storage.ManifestStaticFilesStorage"
    )
STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(BASE_DIR, "static")

GOOGLE_CLIENT_ID = env("OAUTH2_CLIENT_ID", default="")
GOOGLE_CLIENT_SECRET = env("OAUTH2_CLIENT_SECRET", default=None)

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTOCOL", "https")

SESSION_ENGINE = "django.contrib.sessions.backends.signed_cookies"
SESSION_COOKIE_HTTPONLY = True
# SESSION_EXPIRE_AT_BROWSER_CLOSE = True
SESSION_COOKIE_AGE = 3 * 86400
SESSION_SAVE_EVERY_REQUEST = True

MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"

FINEFORMS_WRAPPERS = {"field": "workbench.wrappers.BootstrapFieldWrapper"}
FORM_RENDERER = "django.forms.renderers.TemplatesSetting"


def font(name):
    return os.path.join(BASE_DIR, "stuff", "fonts", name)


NAMESPACE = env("NAMESPACE", required=True)
WORKBENCH = {
    "feinheit": types.SimpleNamespace(
        TITLE="Feinheit",
        SSO_DOMAIN="feinheit.ch",
        CURRENCY="CHF",
        PDF_LANGUAGE="de",
        PDF_COMPANY="Feinheit AG",
        PDF_ADDRESS=" · ".join(
            [
                "Feinheit AG",
                "Fabrikstrasse 54, 8005 Zürich",
                "Effingerstrasse 2, 3011 Bern",
                "www.feinheit.ch",
            ]
        ),
        PDF_VAT_NO="CHE-113.948.417 MWST",
        PDF_OFFER_TERMS=(
            "Bestandteil dieser Offerte sind die zum Zeitpunkt"
            " des Vertragsabschlusses aktuellen Allgemeinen"
            " Geschäftsbedingungen der Feinheit AG."
            " Die jeweils aktuelle Version finden Sie auf www.feinheit.ch/agb/."
        ),
        PDF_INVOICE_PAYMENT=(
            "Wir bedanken uns für die Überweisung des Betrags mit Angabe"
            " der Referenznummer %(code)s bis zum %(due)s"
            " auf ZKB Konto IBAN CH52 0070 0114 8022 0855 1."
        ),
        FONTS={
            "regular": font("ZuricBTLig.ttf"),
            "bold": font("ZuricBTBol.ttf"),
            "italic": font("ZuricBTLigIta.ttf"),
            "bolditalic": font("ZuricBTBolIta.ttf"),
        },
        URL="https://workbench.feinheit.ch",
        FEATURES={
            "bookkeeping": {"jg@feinheit.ch", "mk@feinheit.ch"},
            "campaigns": True,
            "controlling": True,
            "deals": True,
            "foreign_currencies": False,
            "glassfrog": True,
            "labor_costs": False,
            "planning": True,
            "skip_breaks": False,
            "working_time_correction": {"jg@feinheit.ch", "mk@feinheit.ch"},
        },
        READ_WRITE_ADMIN=False,
    ),
    "dbpag": types.SimpleNamespace(
        TITLE="DBAG",
        SSO_DOMAIN="diebruchpiloten.com",
        CURRENCY="CHF",
        PDF_LANGUAGE="de",
        PDF_COMPANY="Die Bruchpiloten AG",
        PDF_ADDRESS=(
            "Die Bruchpiloten AG · Fabrikstrasse 54 · 8005 Zürich"
            " · diebruchpiloten.com"
        ),
        PDF_VAT_NO="CHE-239.647.366 MWST",
        PDF_OFFER_TERMS=(
            "Bestandteil dieser Offerte sind die zum Zeitpunkt"
            " des Vertragsabschlusses aktuellen Allgemeinen"
            " Geschäftsbedingungen der Die Bruchpiloten AG."
        ),
        PDF_INVOICE_PAYMENT=(
            "Wir bedanken uns für die Überweisung des Betrags mit Angabe"
            " der Referenznummer %(code)s bis zum %(due)s"
            " auf ZKB Konto IBAN CH78 0070 0110 0070 4877 9."
        ),
        FONTS={
            "regular": font("HelveticaNeueLight.ttf"),
            "bold": font("HelveticaNeueBold.ttf"),
            "italic": font("HelveticaNeueLightItalic.ttf"),
            "bolditalic": font("HelveticaNeueBoldItalic.ttf"),
        },
        URL="https://workbench.diebruchpiloten.com",
        FEATURES={
            "bookkeeping": True,
            "campaigns": False,
            "controlling": True,
            "deals": False,
            "foreign_currencies": False,
            "glassfrog": False,
            "labor_costs": False,
            "planning": False,
            "skip_breaks": False,
            "working_time_correction": True,
        },
        READ_WRITE_ADMIN=False,
    ),
    "bf": types.SimpleNamespace(
        TITLE="Blindflug",
        SSO_DOMAIN="blindflugstudios.com",
        CURRENCY="CHF",
        PDF_LANGUAGE="de",
        PDF_COMPANY="Blindflug Studios AG",
        PDF_ADDRESS=(
            "Blindflug Studios AG · Fabrikstrasse 54 · 8005 Zürich"
            " · blindflugstudios.com"
        ),
        PDF_VAT_NO="CHE-155.657.233 MWST",
        PDF_OFFER_TERMS=(
            "Bestandteil dieser Offerte sind die zum Zeitpunkt"
            " des Vertragsabschlusses aktuellen Allgemeinen"
            " Geschäftsbedingungen der Blindflug Studios AG."
        ),
        PDF_INVOICE_PAYMENT=(
            "Wir bedanken uns für die Überweisung des Betrags mit Angabe"
            " der Referenznummer %(code)s bis zum %(due)s"
            " auf Konto IBAN CH31 0687 7705 0560 9509 7 bei der"
            " Zürcher Landbank AG."
        ),
        FONTS={
            "regular": font("HelveticaNeueLight.ttf"),
            "bold": font("HelveticaNeueBold.ttf"),
            "italic": font("HelveticaNeueLightItalic.ttf"),
            "bolditalic": font("HelveticaNeueBoldItalic.ttf"),
        },
        URL="https://workbench.blindflugstudios.com",
        FEATURES={
            "bookkeeping": {
                "moritz@blindflugstudios.com",
                "jeremy@blindflugstudios.com",
                "frederic@blindflugstudios.com",
                "mk@feinheit.ch",
            },
            "campaigns": False,
            "controlling": {
                "moritz@blindflugstudios.com",
                "jeremy@blindflugstudios.com",
                "frederic@blindflugstudios.com",
                "mk@feinheit.ch",
            },
            "deals": False,
            "foreign_currencies": True,
            "glassfrog": False,
            "labor_costs": {
                "moritz@blindflugstudios.com",
                "jeremy@blindflugstudios.com",
                "frederic@blindflugstudios.com",
                "mk@feinheit.ch",
            },
            "planning": True,
            "skip_breaks": True,  # For now
            "working_time_correction": {
                "moritz@blindflugstudios.com",
                "jeremy@blindflugstudios.com",
                "frederic@blindflugstudios.com",
                "mk@feinheit.ch",
            },
        },
        READ_WRITE_ADMIN=True,
    ),
    "test": types.SimpleNamespace(
        TITLE="Test",
        SSO_DOMAIN="feinheit.ch",
        CURRENCY="CHF",
        PDF_LANGUAGE="de",
        PDF_COMPANY="Feinheit AG",
        PDF_ADDRESS="Feinheit AG · Fabrikstrasse 54 · 8005 Zürich · www.feinheit.ch",
        PDF_VAT_NO="CHE-113.948.417 MWST",
        PDF_OFFER_TERMS=(
            "Bestandteil dieser Offerte sind die zum Zeitpunkt"
            " des Vertragsabschlusses aktuellen Allgemeinen"
            " Geschäftsbedingungen der Feinheit AG."
            " Die jeweils aktuelle Version finden Sie auf www.feinheit.ch/agb/."
        ),
        PDF_INVOICE_PAYMENT=(
            "Wir bedanken uns für die Überweisung des Betrags mit Angabe"
            " der Referenznummer %(code)s bis zum %(due)s"
            " auf ZKB Konto IBAN CH52 0070 0114 8022 0855 1."
        ),
        FONTS={
            "regular": font("ZuricBTLig.ttf"),
            "bold": font("ZuricBTBol.ttf"),
            "italic": font("ZuricBTLigIta.ttf"),
            "bolditalic": font("ZuricBTBolIta.ttf"),
        },
        URL="https://workbench-test.feinheit.ch",
        FEATURES={
            "bookkeeping": True,
            "campaigns": False,
            "controlling": True,
            "deals": True,
            "foreign_currencies": True,
            "glassfrog": False,
            "labor_costs": True,
            "planning": True,
            "skip_breaks": False,
            "working_time_correction": True,
        },
        READ_WRITE_ADMIN=False,
    ),
}[env("NAMESPACE", required=True)]

INTERNAL_IPS = ["127.0.0.1"]

if LIVE:  # pragma: no cover
    CSRF_COOKIE_SECURE = True
    X_FRAME_OPTIONS = "DENY"
    SESSION_COOKIE_SECURE = True
    SECURE_CONTENT_TYPE_NOSNIFF = True
    SECURE_BROWSER_XSS_FILTER = True
    SECURE_SSL_REDIRECT = True
    SECURE_HSTS_SECONDS = 604800  # One week
    SECURE_HSTS_INCLUDE_SUBDOMAINS = True
    SECURE_REDIRECT_EXEMPT = [r"^(create-timestamp|list-timestamps)/"]
else:
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
    WORKBENCH.TITLE = "(debug) {}".format(WORKBENCH.TITLE)

MAILCHIMP_API_KEY = env("MAILCHIMP_API_KEY", warn=True)
MAILCHIMP_LIST_ID = env("MAILCHIMP_LIST_ID", warn=True)
GLASSFROG_TOKEN = env("GLASSFROG_TOKEN", warn=True)

# Fine since it's only used for selected views
CORS_ORIGIN_ALLOW_ALL = True

if DEBUG:  # pragma: no cover
    globals().update(dj_email_url.parse(env("EMAIL_URL", default="console:")))
else:  # pragma: no cover
    globals().update(dj_email_url.parse(env("EMAIL_URL", default="smtp:", warn=True)))

if env("SQL", default=False):  # pragma: no cover
    from django.utils.log import DEFAULT_LOGGING as LOGGING

    LOGGING["handlers"]["console"]["level"] = "DEBUG"
    LOGGING["loggers"]["django.db.backends"] = {
        "level": "DEBUG",
        "handlers": ["console"],
        "propagate": False,
    }

FEATURES = WORKBENCH.FEATURES
TEST_RUNNER = "django_slowtests.testrunner.DiscoverSlowestTestsRunner"
TESTS_REPORT_TMP_FILES_PREFIX = "tmp/slowtests_"
NUM_SLOW_TESTS = 20

BATCH_MAX_ITEMS = 250

if TESTING:  # pragma: no cover
    PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]
    DATABASES["default"]["TEST"] = {"SERIALIZE": False}
    FEATURES = defaultdict(lambda: True)



from django.urls import re_path

from workbench import generic
from workbench.accounts.features import deals_only
from workbench.deals import views
from workbench.deals.forms import DealForm, DealSearchForm, SetStatusForm
from workbench.deals.models import Deal


urlpatterns = [
    re_path(
        r"^$",
        deals_only(
            generic.ListView.as_view(
                model=Deal, search_form_class=DealSearchForm, paginate_by=None
            )
        ),
        name="deals_deal_list",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        deals_only(generic.DetailView.as_view(model=Deal)),
        name="deals_deal_detail",
    ),
    re_path(
        r"^create/$",
        deals_only(generic.CreateView.as_view(form_class=DealForm, model=Deal)),
        name="deals_deal_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        deals_only(generic.UpdateView.as_view(form_class=DealForm, model=Deal)),
        name="deals_deal_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/set-status/$",
        deals_only(
            generic.UpdateView.as_view(
                form_class=SetStatusForm, model=Deal, template_name="modalform.html"
            )
        ),
        name="deals_deal_set_status",
    ),
    re_path(
        r"^(?P<pk>[0-9]+)/add-offer/$",
        deals_only(views.add_offer),
        name="deals_deal_add_offer",
    ),
    re_path(
        r"^(?P<pk>[0-9]+)/remove-offer/$",
        deals_only(views.remove_offer),
        name="deals_deal_remove_offer",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        deals_only(generic.DeleteView.as_view(model=Deal)),
        name="deals_deal_delete",
    ),
]



import datetime as dt
from decimal import Decimal

from django.contrib import messages
from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.projects.models import Service
from workbench.tools.formats import hours_and_minutes
from workbench.tools.models import HoursField, Model, MoneyField, SearchQuerySet
from workbench.tools.urls import model_urls
from workbench.tools.validation import logbook_lock, raise_if_errors


@model_urls
class LoggedHours(Model):
    service = models.ForeignKey(
        Service,
        on_delete=models.PROTECT,
        related_name="loggedhours",
        verbose_name=_("service"),
    )
    created_at = models.DateTimeField(
        _("created at"), default=timezone.now, db_index=True
    )
    created_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        verbose_name=_("created by"),
        related_name="loggedhours_set",
    )
    rendered_on = models.DateField(
        _("rendered on"), default=dt.date.today, db_index=True
    )
    rendered_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        related_name="loggedhours",
        verbose_name=_("rendered by"),
    )
    hours = HoursField(_("hours"), validators=[MinValueValidator(Decimal("0.1"))])
    description = models.TextField(_("description"))

    invoice_service = models.ForeignKey(
        "invoices.Service",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("invoice service"),
    )
    archived_at = models.DateTimeField(_("archived at"), blank=True, null=True)

    class Meta:
        indexes = [models.Index(fields=["-rendered_on"])]
        ordering = ("-rendered_on", "-created_at")
        verbose_name = _("logged hours")
        verbose_name_plural = _("logged hours")

    def __str__(self):
        return "%s: %s" % (self.service.title, self.description)

    @classmethod
    def allow_delete(cls, instance, request):
        if instance.invoice_service_id or instance.archived_at:
            messages.error(request, _("Cannot delete archived logged hours."))
            return False
        if instance.rendered_on < logbook_lock():
            messages.error(request, _("Cannot delete logged hours from past weeks."))
            return False
        return super().allow_delete(instance, request)

    @classmethod
    def get_redirect_url(cls, instance, request):
        if not request.is_ajax():
            return cls.urls["list"] + "?project={}".format(
                instance.service.project_id if instance else ""
            )


class LoggedCostQuerySet(SearchQuerySet):
    def expenses(self, *, user):
        return self.filter(are_expenses=True, rendered_by=user)


@model_urls
class LoggedCost(Model):
    service = models.ForeignKey(
        Service,
        on_delete=models.PROTECT,
        related_name="loggedcosts",
        verbose_name=_("service"),
    )

    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    created_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("created by")
    )
    rendered_on = models.DateField(
        _("rendered on"), default=dt.date.today, db_index=True
    )
    rendered_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        related_name="loggedcosts",
        verbose_name=_("rendered by"),
    )
    cost = MoneyField(_("cost"))
    third_party_costs = MoneyField(
        _("third party costs"),
        blank=True,
        null=True,
        help_text=_("Total incl. tax for third-party services."),
    )
    description = models.TextField(_("description"))

    invoice_service = models.ForeignKey(
        "invoices.Service",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("invoice service"),
    )
    archived_at = models.DateTimeField(_("archived at"), blank=True, null=True)

    are_expenses = models.BooleanField(_("paid from my own pocket"), default=False)
    expense_report = models.ForeignKey(
        "expenses.ExpenseReport",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("expense report"),
        related_name="expenses",
    )

    expense_currency = models.CharField(
        _("original currency"), max_length=3, blank=True
    )
    expense_cost = MoneyField(_("original cost"), blank=True, null=True)

    objects = LoggedCostQuerySet.as_manager()

    class Meta:
        ordering = ("-rendered_on", "-created_at")
        verbose_name = _("logged cost")
        verbose_name_plural = _("logged costs")

    def __str__(self):
        return self.description

    @classmethod
    def allow_delete(cls, instance, request):
        if instance.invoice_service_id or instance.archived_at:
            messages.error(request, _("Cannot delete archived logged cost entries."))
            return False
        if instance.expense_report:
            messages.error(
                request,
                _("Expenses are part of an expense report, cannot delete entry."),
            )
        return super().allow_delete(instance, request)

    @classmethod
    def get_redirect_url(cls, instance, request):
        if not request.is_ajax():
            return cls.urls["list"] + "?project={}".format(
                instance.service.project_id if instance else ""
            )

    def clean_fields(self, exclude):
        super().clean_fields(exclude)
        errors = {}
        expense = (self.expense_currency != "", self.expense_cost is not None)
        if any(expense) and not all(expense):
            if self.expense_currency == "":
                errors["expense_currency"] = _("Either fill in all fields or none.")
            if self.expense_cost is None:
                errors["expense_cost"] = _("Either fill in all fields or none.")
        raise_if_errors(errors, exclude)


@model_urls
class Break(Model):
    user = models.ForeignKey(
        User, on_delete=models.CASCADE, verbose_name=_("user"), related_name="breaks"
    )
    starts_at = models.DateTimeField(_("starts at"))
    ends_at = models.DateTimeField(_("ends at"))

    description = models.TextField(_("description"), blank=True)

    class Meta:
        ordering = ["-starts_at"]
        verbose_name = _("break")
        verbose_name_plural = _("breaks")

    def __str__(self):
        parts = [
            hours_and_minutes(self.timedelta.total_seconds()),
            self.description,
        ]
        return " ".join(filter(None, parts))

    @property
    def timedelta(self):
        return self.ends_at - self.starts_at

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        errors = {}
        if self.starts_at and self.ends_at:

            if self.starts_at.date() != self.ends_at.date():
                errors["ends_at"] = _("Breaks must start and end on the same day.")

            if self.starts_at >= self.ends_at:
                errors["ends_at"] = _("Breaks should end later than they begin.")

        raise_if_errors(errors, exclude)

    @classmethod
    def allow_update(cls, instance, request):
        if instance.user == request.user:
            return True
        messages.error(request, _("Cannot modify breaks of other users."))
        return False

    allow_delete = allow_update



# Generated by Django 2.2.5 on 2019-09-16 09:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("invoices", "0016_auto_20190916_1108")]

    operations = [
        migrations.RunSQL(
            "SET SESSION application_name TO 'Migrations';"
            "UPDATE invoices_invoice SET total_excl_tax=subtotal - discount - down_payment_total;"
            "UPDATE invoices_recurringinvoice SET total_excl_tax=subtotal - discount;"
        )
    ]



import datetime as dt
from decimal import Decimal
from itertools import takewhile

from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import F, Sum, signals
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from authlib.email import render_to_mail

from workbench.accounts.models import User
from workbench.invoices.utils import recurring
from workbench.offers.models import Offer
from workbench.projects.models import Project
from workbench.tools.formats import Z1, hours, local_date_format
from workbench.tools.models import HoursField, Model, SearchQuerySet
from workbench.tools.urls import model_urls
from workbench.tools.validation import raise_if_errors


class PlanningRequestQuerySet(SearchQuerySet):
    def with_missing_hours(self):
        return self.annotate(
            _missing_hours=F("requested_hours") - F("planned_hours")
        ).filter(project__closed_on__isnull=True, _missing_hours__gt=0)

    def maybe_actionable(self, *, user):
        return (
            self.with_missing_hours().filter(receivers=user).select_related("project")
        )


@model_urls
class PlanningRequest(Model):
    project = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        verbose_name=_("project"),
        related_name="planning_requests",
    )
    offer = models.ForeignKey(
        Offer,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("offer"),
        related_name="planning_requests",
    )
    requested_hours = HoursField(
        _("requested hours"), validators=[MinValueValidator(Decimal("0.1"))]
    )
    planned_hours = HoursField(_("planned hours"))
    earliest_start_on = models.DateField(_("earliest start on"))
    completion_requested_on = models.DateField(_("completion requested on"))
    title = models.CharField(_("title"), max_length=100)
    description = models.TextField(_("description"), blank=True)
    receivers = models.ManyToManyField(
        User, verbose_name=_("receivers"), related_name="received_planning_requests"
    )

    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    created_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        verbose_name=_("created by"),
        related_name="sent_planning_requests",
    )
    closed_at = models.DateTimeField(_("closed at"), blank=True, null=True)
    is_provisional = models.BooleanField(_("is provisional"), default=False)

    objects = PlanningRequestQuerySet.as_manager()

    class Meta:
        ordering = ["-pk"]
        verbose_name = _("planning request")
        verbose_name_plural = _("planning requests")

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if self.pk:
            self.planned_hours = (
                self.planned_work.order_by().aggregate(h=Sum("planned_hours"))["h"]
                or Z1
            )
        else:
            self.planned_hours = Z1
        super().save(*args, **kwargs)

    save.alters_data = True

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        errors = {}

        if self.earliest_start_on.weekday() != 0:
            errors["earliest_start_on"] = _("Only mondays allowed.")
        if self.completion_requested_on.weekday() != 0:
            errors["completion_requested_on"] = _("Only mondays allowed.")

        if self.completion_requested_on <= self.earliest_start_on:
            errors["completion_requested_on"] = _(
                "Allow at least one week for the work please."
            )

        raise_if_errors(errors, exclude)

    @cached_property
    def weeks(self):
        return list(
            takewhile(
                lambda x: x < self.completion_requested_on,
                recurring(self.earliest_start_on, "weekly"),
            )
        )

    @cached_property
    def receivers_with_work(self):
        work = {user: [] for user in self.receivers.all()}
        for pw in self.planned_work.select_related("user"):
            work.setdefault(pw.user, []).append(pw)
        return sorted(work.items())

    @property
    def missing_hours(self):
        return self.requested_hours - self.planned_hours

    def html_link(self):
        return format_html(
            '<a href="{}" data-toggle="ajaxmodal">{}</a>', self.get_absolute_url(), self
        )


def receivers_changed(sender, action, instance, pk_set, **kwargs):
    if action == "post_add":
        users = User.objects.filter(pk__in=pk_set)
        render_to_mail(
            "planning/planningrequest_notification",
            {"object": instance, "WORKBENCH": settings.WORKBENCH},
            to=[user.email for user in users],
            cc=[instance.created_by.email],
            reply_to=[instance.created_by.email] + [user.email for user in users],
        ).send(fail_silently=True)


signals.m2m_changed.connect(receivers_changed, sender=PlanningRequest.receivers.through)


@model_urls
class PlannedWork(Model):
    project = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        verbose_name=_("project"),
        related_name="planned_work",
    )
    offer = models.ForeignKey(
        Offer,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        verbose_name=_("offer"),
        related_name="planned_work",
    )
    request = models.ForeignKey(
        PlanningRequest,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        verbose_name=_("planning request"),
        related_name="planned_work",
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    user = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        verbose_name=_("user"),
        related_name="planned_work",
    )
    planned_hours = HoursField(
        _("planned hours"), validators=[MinValueValidator(Decimal("0.1"))]
    )
    title = models.CharField(_("title"), max_length=200)
    notes = models.TextField(_("notes"), blank=True)
    weeks = ArrayField(models.DateField(), verbose_name=_("weeks"))

    class Meta:
        ordering = ["-pk"]
        verbose_name = _("planned work")
        verbose_name_plural = _("planned work")

    def __str__(self):
        return "{} ({})".format(self.title, hours(self.planned_hours))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._update_request_ids = {self.request_id}

    def _update_requests(self):
        self._update_request_ids.add(self.request_id)
        self._update_request_ids.discard(None)
        if self._update_request_ids:
            for request in PlanningRequest.objects.filter(
                id__in=self._update_request_ids
            ):
                request.save()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self._update_requests()

    save.alters_data = False

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        self._update_requests()

    delete.alters_data = False

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        errors = {}

        if self.weeks:
            no_mondays = [day for day in self.weeks if day.weekday() != 0]
            if no_mondays:
                errors["weeks"] = _("Only mondays allowed, but field contains %s.") % (
                    ", ".join(local_date_format(day) for day in no_mondays),
                )

        raise_if_errors(errors, exclude)

    @property
    def pretty_from_until(self):
        return "{} – {}".format(
            local_date_format(min(self.weeks)),
            local_date_format(max(self.weeks) + dt.timedelta(days=6)),
        )

    @property
    def ranges(self):
        def _find_ranges(weeks):
            start = weeks[0]
            maybe_end = weeks[0]
            for day in weeks[1:]:
                if (day - maybe_end).days == 7:
                    maybe_end = day
                else:
                    yield start, maybe_end
                    start = maybe_end = day

            yield start, maybe_end

        for from_, until_ in _find_ranges(self.weeks):
            yield {
                "from": from_,
                "until": until_,
                "pretty": "{} – {}".format(
                    local_date_format(from_),
                    local_date_format(until_ + dt.timedelta(days=6)),
                ),
            }

    @property
    def pretty_planned_hours(self):
        return _("%(planned_hours)s in %(weeks)s weeks (%(per_week)s per week)") % {
            "planned_hours": hours(self.planned_hours),
            "weeks": len(self.weeks),
            "per_week": hours(self.planned_hours / len(self.weeks)),
        }



import datetime as dt

from django import forms
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.accounts.features import FEATURES
from workbench.accounts.models import User
from workbench.awt.models import Absence
from workbench.tools.forms import Form, ModelForm, Textarea, WarningsForm, add_prefix
from workbench.tools.xlsx import WorkbenchXLSXDocument


class AbsenceSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    u = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    reason = forms.ChoiceField(
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        choices=[("", _("All reasons"))] + Absence.REASON_CHOICES,
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["u"].choices = User.objects.choices(
            collapse_inactive=False, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("u") == -1:
            queryset = queryset.filter(user=self.request.user)
        elif data.get("u"):
            queryset = queryset.filter(user=data.get("u"))
        if data.get("reason"):
            queryset = queryset.filter(reason=data.get("reason"))
        return queryset.select_related("user")

    def response(self, request, queryset):
        if (
            request.GET.get("export") == "xlsx"
            and request.user.features[FEATURES.CONTROLLING]
        ):
            xlsx = WorkbenchXLSXDocument()
            xlsx.table_from_queryset(queryset)
            return xlsx.to_response("absences.xlsx")


@add_prefix("modal")
class AbsenceForm(ModelForm, WarningsForm):
    user_fields = default_to_current_user = ("user",)

    class Meta:
        model = Absence
        fields = ["user", "starts_on", "ends_on", "days", "reason", "description"]
        widgets = {"description": Textarea({"rows": 2})}

    def __init__(self, *args, **kwargs):
        initial = kwargs.setdefault("initial", {})
        request = kwargs["request"]
        for field in ["user", "starts_on", "ends_on", "days", "reason", "description"]:
            if value := request.GET.get(field):
                initial[field] = value

        super().__init__(*args, **kwargs)
        self.fields["days"].help_text = format_html(
            '<a href="#" data-hours-button="{hours}">{hours}</a>',
            hours=_("Enter hours"),
        )
        if not self.request.user.features[FEATURES.WORKING_TIME_CORRECTION]:
            self.fields["reason"].choices = [
                choice
                for choice in self.fields["reason"].choices
                if choice[0] != Absence.CORRECTION
            ]

    def clean(self):
        data = super().clean()
        if (starts_on := data.get("starts_on")) and (user := data.get("user")):
            today = dt.date.today()
            if user.enforce_same_week_logging and starts_on.year < today.year:
                self.add_error(
                    "starts_on", _("Creating absences for past years is not allowed.")
                )

            elif starts_on.year < today.year:
                self.add_warning(
                    _(
                        "Creating absences for past years has effects on financial"
                        " statements, carry-forward entries etc. Are you sure?"
                    ),
                    code="past-years-absences",
                )

            elif (starts_on - today).days > 366:
                self.add_warning(
                    _(
                        "Impressive planning skills or wrong date?"
                        " Absence starts in more than one year."
                    ),
                    code="impressive-planning-skills",
                )

        return data
